/*****************************************************************************/
//    GArray.h
/*****************************************************************************/

#ifndef _GArray_H_
#define _GArray_H_

#ifndef GENERICH
#include <generic.h>
#endif

/*****************************************************************************/

extern int corC_NewGArraySize, corC_DeleteGArraySize;

/*****************************************************************************/
/*
    The following must be used to declare any variable of the generic
    type, G(rowing)Array
*/

#define GArray(type) CONCAT_ID2(type,_garray)
#define ceil_div(size, incr) ((size+incr-1)/incr)

/*****************************************************************************/
/*
    The following function must be used to declare GArray(type)'s definition.
    It must be included at most once in a program, so #ifdef as appropriate
    may need to be used.
*/

#define GArrayDeclare(type)            \
    class GArray(type) {            \
    type **corarray;                \
    int size;                \
    int _limit;                \
    int incr;                 \
    public:                    \
        GArray(type) (int size =128*128, int incr = 128, int tlimit=-1);    \
        /* type &operator[](int i); */            \
            inline type & operator[](int i)            \
            {                            \
        ASSERT(corarray!=NULL && i<_limit && i >= 0);    \
        if (i >= size) resize(2*size);            \
        if (! corarray[i/incr]) {                 \
                corarray[i/incr] = new type[incr];        \
            corC_NewGArraySize += incr*sizeof(type);    \
        }                        \
        return ( corarray[i/incr][i%incr] );        \
            }                            \
        int limit() {return _limit;}                \
        int resize(int newsize);                 \
        GArray(type) *make_copy();                \
        int get_next(int i);                \
        ~GArray(type)();            \
    };    

/*
        CONCAT_ID2(~,GArray(type)());            \
*/


/*
 * The following function must be used to define the implementation of 
 * GArray(type).  It must occur only once, else the implementation would be 
 * multiply defined.
 */

#define GArrayImplement(type)                \
    GArray(type)::GArray(type)(int tsize, int inc, int tlimit)     \
    {                            \
    size = tsize;                    \
    if (tlimit < 0) _limit = size;            \
    else _limit = tlimit;                \
    incr = inc;                    \
    if (size<incr) incr = size;            \
    if (tsize <= 0) corarray = NULL;             \
    else  corarray = new type *[ceil_div(tsize,inc)];    \
    corC_NewGArraySize += ceil_div(tsize,inc)*sizeof(type *);    \
    for(int i=0;i<ceil_div(tsize,inc);i++) corarray[i]=NULL;      \
    }                                \
                                \
    int GArray(type)::resize(int newsize)             \
    {                                \
    int i;                            \
    if(newsize <= size) { size = newsize; return 0;}    \
    if (_limit < newsize) _limit = newsize;            \
    type **newarray = new type *[ceil_div(newsize,incr)];    \
    for(i=0;i<ceil_div(size,incr);i++)             \
        newarray[i]=corarray[i];                \
    for(i=ceil_div(size,incr);i<ceil_div(newsize,incr);i++) \
        newarray[i]=NULL;                \
    corC_NewGArraySize += (newsize - size)*sizeof(type *);    \
    size = newsize;                        \
    delete corarray;                        \
    corarray = newarray;                    \
    return 0;                        \
    }                            \
                            \
    GArray(type) * GArray(type)::make_copy()        \
    {                            \
    int i,j;                    \
    GArray(type) *copy = new GArray(type);        \
                            \
    copy->size = size;                \
    copy->_limit = _limit;                \
    copy->incr = incr;                \
    if(size<=0)    copy->corarray = NULL;        \
    else copy->corarray = new type *[ceil_div(size,incr)];    \
    corC_NewGArraySize += size * sizeof(type *);    \
    for(i=0;i<ceil_div(size,incr);i++){        \
        if(corarray[i]) {                \
            copy->corarray[i] = new type[incr];    \
            for (j=0;j<incr;j++)            \
            copy->corarray[i][j] = corarray[i][j];    \
        corC_NewGArraySize += incr*sizeof(type);     \
        }                    \
        else copy->corarray[i] = NULL;            \
        }                        \
    return (copy) ;                    \
    }                            \
                            \
    int GArray(type)::get_next(int i)            \
    {                            \
    if ( !corarray ) return(-1);            \
    for(i++;i<size;){                \
       if (! corarray[i/incr])                \
        i = ((i/incr)+1)*incr;            \
       else break;                    \
       }                        \
    return(i);                    \
     }                            \
                            \
     GArray(type)::~GArray(type)()            \
     {                            \
    if (!corarray) return;                \
    for(int i=0; i< ceil_div(size,incr);i++) {    \
        if(corarray[i]) {                \
           delete [incr] corarray[i];            \
           corC_DeleteGArraySize += incr*sizeof(type);\
        }                        \
    }                        \
    delete [ceil_div(size,incr)] corarray ;        \
    corC_DeleteGArraySize += ceil_div(size,incr)*sizeof(type *); \
     }
            
/*****************************************************************************/

#endif _GArray_H_

