// *************************************************************************
//    ParseUtils.h
// *************************************************************************

#ifndef _ParseUtils_H_
#define _ParseUtils_H_

#include <stdio.h>

#include "Arg.h"
#include "FuncArg.h"
#include "CArg.h"
#include "Annotations.h"

// *************************************************************************

class corParserStruct;
class corModuleInfo;
class corPredAnnotations;
class corClause;

// *************************************************************************

typedef struct corarg_link {
    struct corarg_link *prev;
    corCArg arg;
} Arg_Link;

typedef struct corarg_chain {
    Arg_Link *last;
    int count;
    struct corarg_chain *pushed;
} Arg_Chain;

// *************************************************************************

typedef struct corparameter_link {
    struct corparameter_link *prev;
    corCArg *parameter_slot;
} Parameter_Link;

struct corparameter_chain {
    Parameter_Link *last;
    int count;

    corparameter_chain() { last = NULL; count = 0; }
};

// *************************************************************************

struct corset_stack {
    struct corset_stack *prev;
    ArgPtr info;
};

typedef struct corset_stack corset_stack;

// *************************************************************************
// defined in Gram.y

extern corSymbol ** corbuiltin_ops[];
extern int corBuiltin_Arith_Start;
extern int corBuiltin_Convertible_Start;

// *************************************************************************

extern void corpush_arg_chain(corParserStruct& corparserStruct);
extern void corappend_arg(corParserStruct& corparserStruct, corCArg* a);
extern void corappend_arg(corParserStruct& corparserStruct, corArg *a);
extern void corappend_parameter(corParserStruct& corparserStruct, corCArg *a, Parameter_Link*&);
extern void corpop_arg_chain(corParserStruct& corparserStruct, corArgList& args);

extern void corpush_cur_set(corParserStruct& corparserStruct, ArgPtr set);
extern void corpop_cur_set(corParserStruct& corparserStruct);
extern void corinsert_cur_set(corParserStruct& corparserStruct, corCArg* element);

extern corCArg* corcreate_list(corParserStruct& corparserStruct, ArgPtr cdr);
extern corCArg* corcreate_list(corParserStruct& corparserStruct, corCArg* cdr);

extern corArg* corcreate_functor(corParserStruct& corparserStruct, corCArg*);
extern ArgPtr corcreate_grouping(corParserStruct& corparserStruct);
extern ArgPtr corcreate_multi_grouping(corParserStruct& corparserStruct);
extern ArgPtr corcreate_nested_multiset();

extern corVarArg* corpull_out_arith_expr(corParserStruct&);

extern int corrule_var_enter(corParserStruct& corparserStruct, Name s);
inline int corrule_var_enter(corParserStruct& corparserStruct, corCArg* c) {
  if (c)
    return corrule_var_enter(corparserStruct, c->SymbolP());
  else{
   CORAL_error(COR_FATAL_ERROR, "carg == 0","rule_var_enter(.. CArg*)");
   return 0;
  }
 };

extern corrule_var_lookup(corParserStruct& corparserStruct, corCArg*);
extern int corsip_var_lookup(corParserStruct& corparserStruct, corCArg*);

extern void corAddExportInfo(corParserStruct& corparserStruct, corCArg*);
extern void corAddExportInfo(corParserStruct& corparserStruct, corFuncArg*);

extern void corcomplete_rule(corClause *cur_rule, corParserStruct& corparserStruct);

extern void COR_set_save_module(corModuleInfo&, int val);
extern void COR_set_supplementary_magic(corModuleInfo&, int val);
extern void COR_set_ordinary_magic(corModuleInfo&, int val);
extern void COR_set_no_magic(corModuleInfo&);
extern void COR_set_sup_magic_indexing(corModuleInfo&, int val);
extern void COR_set_predicate_seminaive(corModuleInfo&, int val);
extern void COR_set_check_subsum(corModuleInfo&, int val);
extern void COR_set_multisets(corModuleInfo&, int val);
extern void COR_set_single_scc(corModuleInfo&, int val);
extern void COR_set_convert_functions(corModuleInfo&, int val);
extern void COR_set_exist_opt(corModuleInfo&, int val);
extern void COR_set_use_factoring(corModuleInfo&, int val);
extern void COR_set_use_factormagic(corModuleInfo&, int val);
extern void COR_set_use_ordsearch(corModuleInfo&, int val);
extern void COR_set_return_unify(corModuleInfo&, int val);
extern void COR_set_pipelining(corModuleInfo&, int val);
extern void COR_set_index_deltas(corModuleInfo&, int val);
extern void COR_set_eager_eval(corModuleInfo&, int val);
extern void COR_set_monotonic(corModuleInfo&, int val);
extern void COR_set_non_ground_facts(corModuleInfo&, int val);
extern void COR_set_rewriting(corModuleInfo&, int val);
extern void COR_set_preprocessing(corModuleInfo&, int val);
extern void COR_set_disk_resident(corModuleInfo&, int);
extern void COR_set_nested_scc_eval(corModuleInfo&, int);


// Functions in annotations.C
//    REMOVE

/*
extern void add_make_index_annotation(PredAnnotations *&, Symbol *pred,
        Symbol *adorn_string, ArgList *al1, ArgList *al2);
extern void add_allowed_adornment(ParserStruct&, Name predname, ArgList *arg);
extern void add_multiset_annotation(ParserStruct&, Symbol *pred,
        ArgList *al1, int check_subsum);
extern void add_diskrel_annotation(ParserStruct&, Symbol *pred,
        Symbol *adorn_string,ArgList *al1);
extern void add_agg_sel_annotation(ParserStruct&, Symbol *pred,
        Symbol *adorn_string, ArgList *al1, ArgList *al2, Arg *op_arg);
extern void add_index_delta_annotation(ParserStruct&, Symbol *pred,
        ArgList *al1);
extern void add_prioritize_annotation(ParserStruct&, Symbol *pred,
        Symbol *adorn_string, ArgList *al1, Arg *op_arg);
*/


#endif _ParseUtils_H_

