/*****************************************************************************/
//    Relation.h
/*****************************************************************************/

#ifndef _Relation_H_
#define _Relation_H_

#include "RelationMisc.h"

#include "Arg.h"
#include "SymbolArg.h"
#include "AggSelInfo.h"
//#include "PrioritizeInfo.h"
#include "PredAnnotations.h"
//#include "StorageRelation.h"
//#include "Interface.h"
#include "RMark.h"
#include "Tuple.h"

/*****************************************************************************/

class corStorageRelation;
class corPrioritizeInfo;
class corTupleIterator;

// from Interface.h (move to here?)
extern void delete_tuple (corRelation *rel, corTuple *tuple);
extern void update_tuple (corTuple *tuple, int arg_pos, corArg *new_val);
extern void update_tuple (corRelation *rel, corTuple *old_tuple, corTuple *new_tuple);

/*****************************************************************************/

class corRelation : public corArg {
protected:
    int _arity;
public:
    Name name;
    RelationOptions local_options; 
    //unsigned int  check_subsum : 1;        // If set, check for subsumption when
    // inserting tuples. 
    char check_subsum;
    //unsigned int delta_indexed : 1;
    //unsigned int single_success : 1;
    //unsigned int has_release : 1;        // specifies if the relation has a
    
    char delta_indexed, single_success, has_release;
    // release() method
    int count;                            // number of facts in relation
    unsigned int ref_count : 8;            // how many databases it's open in
    unsigned int scan_count : 8;        // how many scans are open on it

    HashVal _hash;                        // sufficient for inequality testing
    HashVal _eq_id;                        // sufficient for equality testing
    corAggSelInfo *agg_sel_info;            // Info on aggregate selections on reln
    corPrioritizeInfo *priority_info;        // Prioritization info on this relation

    RelationKind r_kind;

    //    Ordered Search Fields
    short ordsearch;
    corStorageRelation *for_ordsearch;
    // used for subsumption checking in ordered search
    corStorageRelation *mp_done_reln;
    // non-NULL only if (r_kind == COR_R_MAGIC) && (ordsearch)

    //    Aggregate selection related methods
    virtual void add_agg_sel_anno(corPredAnnotations *) {};
    virtual void add_agg_sel_info(corAggSelInfo *) {};

    //    Prioritize related methods
    void add_prioritize_info(corPredAnnotations *anno);
    virtual int add_hidden_relation(corPrioritizeInfo *info);
    // The above adds a new hidden relation
    virtual int insert_in_hidden_relation(corTuple *tuple);
    virtual int pop_from_hidden_relation();
    // Pop a tuple from the hidden relation and add it
    //    to visible part.
    virtual int pop_all_from_hidden_relation();
    // Pop all tuples from the hidden relation, and
    // add them to visible part.
    
    //    abstract relation methods
    int arity() const { return _arity; }
    virtual void print(corBindEnv*, FILE *file) const;
    virtual void print(corBindEnv*, FILE *file, char *) const;
    virtual void sprint(char* str, int* pos, corBindEnv *context = NULL) const;

    virtual void print_name(FILE *file);
    virtual int isConstant();
    virtual int tuple_count() { return count; }
    void printon(FILE *file) const;
    void printon(FILE *file, char *) const;
    virtual void print_facts(FILE *file, corArgList *arglist = NULL) const ;

    //    get_next_tuple(): returns tuple if there is one (Builtin rels etc
    //    return NULL).  Returns status via tupleit->bindenv->no_match()
    virtual corTuple* get_next_tuple(corTupleIterator&) = 0;
    virtual corBindEnv* get_next(corTupleIterator&) = 0;

    virtual int insert_tuple(corTuple *tuple) = 0;
    virtual int insert_new(corArgList& args, corBindEnv* env,
        corBindEnv *dont_rename_env = NULL, corTuple *parent_goal = NULL,
        int dont_rename = 0) = 0; 

    inline int insert_new(corArg** args, corBindEnv* env,
        corBindEnv *dont_rename_env = NULL, corTuple *parent_goal = NULL,
        int dont_rename = 0)
    {
            return insert_new(*(corArgList*)args, env, dont_rename_env,
                parent_goal, dont_rename);
    }

    inline int insert_new(corTuple *tuple, corBindEnv *dont_rename_env = NULL,
        corTuple *parent_goal = NULL, int dont_rename = 0)
    {
        return insert_new(tuple->args(), tuple->bindenv, dont_rename_env,
            parent_goal, dont_rename);
    }

    virtual void empty_relation(int deleteTuples = 0) {
        deleteTuples = 0; /*dummy*/
    };

    // Empty tuples out of the relation.  Delete the tuples
    // only if the deleteTuples arg is not 0
    virtual int tuple_delete(corTuple *tuple) {
        delete_tuple(this, tuple); 
        return 1;
    }

    virtual int tuple_insert(corTuple *tuple) {
        return this->insert_new(tuple->args(), tuple->bindenv);
    }
    int tuple_update(corTuple *tuple_old, corTuple *tuple_new) {
        update_tuple(this, tuple_old, tuple_new); 
        return 1;
    }

    //    Release state associated with tuple iterator
    virtual void release(corTupleIterator&) { };

    virtual int add_index(corArgList *)                    { return 0; };
    virtual int add_index(corArgList *, int , corArgList *)    { return 0;};
    virtual corRMark * getMark()                            {return NULL ;};
    virtual void freeMark(corRMark *, int hide_em=0)        { hide_em = 0;} ;

    // THESE SHOULD BE IN STORAGERELATION !!!
    virtual HashVal hash(corBindEnv *env = NULL);
    // compute a hash value for a nested set
    virtual int equals(corArg *arg2);
    virtual int raw_equals(corRelation *rel);
    virtual int contained_in(corRelation *rel);
    virtual unsigned int cardinality(corTuple *tuple);

    corRelation( int arity, int delta_indexed = -1 );
    corRelation();
};

/*****************************************************************************/

#endif _Relation_H_

