/*****************************************************************************/
//    RelationMisc.h
/*****************************************************************************/

#ifndef _RelationMisc_H_
#define _RelationMisc_H_

#include "Arg.h"
#include "Unify.h"
#include "SymbolArg.h"
#include "CArg.h"

/*****************************************************************************/

class corStorageRelation;
class corGenericIndex;
class corAggSelInfo;

/*****************************************************************************/
//    Abstract interface for handling "generic" Relations, independent
//    of concrete implementation (linked list, database, builtin, etc).
//    To add further defines, just number them with successive powers of 2 

//    The following two can only be used within methods of GenericRel (or 
//    methods of subclasses of GenericRel) 
#define REL_DISPLAY_NEW_INSERTIONS 1    // Display non-subsumed insertions
#define REL_DISPLAY_INSERTIONS 2        // Display all attempted insertions

//    The following can only be used within methods of GenericIndex (or 
//    methods of subclasses of GenericIndex) 
#define REL_DISPLAY_INDEXOPS   4        // Display calls to index operations

#define REL_PROFILE_STATS  8            // Generate profile statistics

#define REL_EXPLAIN 16                    // Turn browser on(dump)

#define USING_RETURN_UNIFY (corCurrModuleInfo->UseReturnUnify)

/*****************************************************************************/

#define DISPLAY_INSERTIONS  \
       ((corexEnv->GlobalRelationOptions | this->local_options) & REL_DISPLAY_INSERTIONS)
#define DISPLAY_NEW_INSERTIONS  \
       ((corexEnv->GlobalRelationOptions | this->local_options) & REL_DISPLAY_NEW_INSERTIONS)
                
#define DISPLAY_INDEXOPS  \
    ((corexEnv->GlobalRelationOptions | this->relation->local_options) & REL_DISPLAY_INDEXOPS)
                
/*****************************************************************************/
//    iterator.ipos is used to keep track of where were are in the scan
//    of the hash index.  The possible values are:
#define Scanning_Init  0     // Initial state
#define Scanning_Known 1     // scanning known bucket
#define Scanning_Vars  2     // scanning from var_chain
#define Scanning_New   3     // scanning from delta chain
#define Scanning_Done  4     // done
#define Scanning_Hash  5

/*****************************************************************************/
//    For i in [Scanning_Hash .. Scanning_Hash+hash_size-1] means
//    that we are scanning all hash buckets, and are currently at
//    bucket i-Scanning_Hash 

typedef int RelationOptions; // List of possible options follows.

/*****************************************************************************/

enum RelationKind {
    /* the next three are the only possible rkind values possible for
         actual materialized relations. Only these rkinds need to be
         checked for at run time
     */
    COR_R_MAGIC,
    COR_R_SUPPLEMENTARY,
    COR_R_ANSWER,
    COR_R_DERIVED,
    COR_R_EXODUS,
    COR_R_OTHER
};

/*****************************************************************************/

extern int corRelCount;

/*****************************************************************************/

void corgroup_dump(corBindEnv * new_env, corArgList * new_arglist, corRelation *rel,
    char *old_arglist = NULL);

RelationKind corfind_r_kind(Name name);
int corhas_grouping_args(corArgList &args);

corRelation* corfind_local_or_external_relation(Name name, int arity);
corRelation* corfind_external_relation(Name name, int arity);
corRelation* corFind_External_Relation(Name name, int arity);

corRelation* find_relation(char * name, int arity);
corRelation* find_relation(Name name, int arity);
corRelation* find_relation(corCArg c, int arity);

corRelation* cormake_relation(Name name, int arity, int delta_indexed = -1);
corRelation* cormake_relation(char *name, int arity, int delta_indexed = -1);
corStorageRelation* cormake_local_relation(Name name, int arity, int delta_indexed,
    int set_subsum = -1);

void correclaim_relation(corStorageRelation *rel);
void correclaim_index(corGenericIndex *index, RelationKind r_kind);

int corduplicate_agg_sel_anno(corStorageRelation *rel, corAggSelInfo *newinfo);

/*****************************************************************************/

#endif _RelationMisc_H_

