/*****************************************************************************/
//    StorageManager.h
/*****************************************************************************/

#ifndef _StorageManager_H_
#define _StorageManager_H_

/*****************************************************************************/

class corStorageManager {
private:
    corCatalog catalog;

    // Maybe this should have two buffer groups -- an MRU and an LRU
    corBufferGroup buf;
    corTransaction trans;
    VOLID _volid;
    VOLID _temp_volid;

public:
    corStorageManager()    { INVALIDATE_VOLID(_volid); }
    ~corStorageManager()    { shutdown(); }
    void init();

    void shutdown();
    int bufGroup()        { return buf.bufGroup; }
    VOLID volid()        { return _volid; }
    VOLID tempvolid()    { return _temp_volid; }

    // Transaction handling functions
    void startTrans()    { trans.start() ; }
    int commitTrans()    { return trans.commit() ;}
    void abortTrans()    { trans.abort() ;}

    // Catalog handling functions

    // Add a file to the catalog
    int addFile(Name name, int arity, corFileDescStruct *obj);

    // Check if a file is in the catalog
    int lookupFile(Name name, int arity, OID* oid);

    // Delete a file from the catalog
    int deleteFile(Name name, int arity);

    // List all files/schemas
    void listFiles(FILE *outf);

    // Object handling functions

    // Retrieve an object specified by oid. Note that the object is PINNED
    // in the buffer pool !
    int retrieveObject(OID *oid, USERDESC ** u_desc);

    // Release (unpin) an object from the buffer pool.
    int releaseObject(USERDESC *u_desc);

    // Destroy an object. If bufGroup is 0, the object is in the catalog file.
    int destroyObject(OID *oid, int bufGroup);
};

/*****************************************************************************/

extern corStorageManager SM;

/*****************************************************************************/

#endif _StorageManager_H_

