#include <fstream.h>
#include <iomanip.h>
#include "RdbDatabase.h"
#include "RdbCursor.h"
#include "RdbCompilableDML.h"


RdbDatabase::RdbDatabase(const String& newDbName,
			 RdbSystem& newSystem,
			 const String& newServerName)
: system(newSystem),
  dbName(newDbName),
  serverName(newServerName)
#if defined(DO_RDB_LOGGING)
     , dblog(NULL,1,0)
//     , dblog(new ofstream(String("/tmp/dblog.")+newDbName,ios::app),1,1)
#endif
{
#if defined(DO_RDB_LOGGING)
    ofstream* os = new ofstream(String("/tmp/dblog.")+newDbName,ios::app);
    os->setf(ios::unitbuf);
    dblog.SetStream(os,1);
    dblog.Enable();
#endif
    connected = FALSE;
    system.RegisterDatabase(this);
    RDBLOG(dblog, "Registered database" << endl);
}


RdbDatabase::~RdbDatabase()
{
    if( connected )
      Disconnect();
    system.DeregisterDatabase(this);
    RDBLOG(dblog, "Deregistered database" << endl);
    // and hope that nothing else references this...
}


void RdbDatabase::RegisterCompilableDML( RdbCompilableDML& CompilableDML )
{
    compilableDMLList.Append( &CompilableDML );
}


void RdbDatabase::DeregisterCompilableDML( RdbCompilableDML& CompilableDML )
{
    for( int i = compilableDMLList.Size() -1; i >= 0 ; i-- )
      if( compilableDMLList[i] == &CompilableDML ) {
	  compilableDMLList.Delete(i);
	  return;
      }
    assert(0);			// better be found!
}


void RdbDatabase::InvalidateCompilableDMLs()
{
    for( int i = 0 ; i < compilableDMLList.Size() ; i++ ) {
	compilableDMLList[i]->Invalidate();
    }
}


void RdbDatabase::RegisterCursor( RdbCursor& Cursor )
{
    cursorList.Append( &Cursor );
}


void RdbDatabase::DeregisterCursor( RdbCursor& Cursor )
{
    for( int i = cursorList.Size() -1; i >= 0 ; i-- )
      if( cursorList[i] == &Cursor ) {
	  cursorList.Delete(i);
	  return;
      }
    assert(0);			// better be found!
}


void RdbDatabase::InvalidateCursors()
{
    for( int i = 0 ; i < cursorList.Size() ; i++ ) {
	cursorList[i]->Invalidate();
    }
}


RdbError::ErrorCode RdbDatabase::Connect()
{
    if( connected )
      return NO_ERROR;
    RDBLOG(dblog, "Connecting to " << dbName << endl);
    return VConnect();
}


RdbError::ErrorCode RdbDatabase::Disconnect()
{
    if( !connected )
      return NO_ERROR;
    SetFocus(); 
    RDBLOG(dblog, "Disconnecting from " << dbName << endl);
    return VDisconnect();
}


RdbError::ErrorCode RdbDatabase::ImmediateSql(const String& sql)
{
    ErrorCode rc = NO_ERROR;
    SetFocus(); 
    if( !connected ) {
	rc = Connect();
	if( rc )
	  return rc;
    }
    RDBLOG(dblog, "Executing Immediate sql: " << sql << endl);
    return VImmediateSql(sql);
}


RdbError::ErrorCode RdbDatabase::Commit()
{
    if( !connected )
      return NO_ERROR;
    SetFocus(); 
    RDBLOG(dblog, "Commit" << endl);
    return VCommit();
}


RdbError::ErrorCode RdbDatabase::Rollback()
{
    if( !connected )
      return NO_ERROR;
    SetFocus(); 
    RDBLOG(dblog, "Rollback" << endl);
    return VRollback();
}


String RdbDatabase::GetErrorMessage()
{
    SetFocus(); 
    RDBLOG(dblog, "Getting error message" << endl);
    return VGetErrorMessage();
}


RdbError::ErrorCode RdbDatabase::SetFocus()
{
    ErrorCode rc = NO_ERROR;
    if( !connected ) {
	Connect();
    }
    if( !system.IsCurrentDatabase(this) ) {
	RDBLOG(dblog, "Setting focus to me" << endl);
	system.SetFocusToDatabase(this);
	VSetFocus();
    }
    return rc;
}
