/*****************************************************************************/
//	Arg.C
/*****************************************************************************/

#include <stdlib.h>
#include <string.h>

#include "Arg.h"
#include "DontCareArg.h"
#include "Globals.h"
#include "Term.h"
#include "ArgList.h"

/*****************************************************************************/

extern corArrayBindEnv corUnusedEnv;

/*****************************************************************************/
void corArg::print(corBindEnv *, FILE *) const
{
}

/*****************************************************************************/
void corArg::print(corBindEnv *, FILE *, char *) const
{
}

/*****************************************************************************/
void corArg::printon(FILE *, char *) const
{
}

/*****************************************************************************/
corCArg corArg::simplify(corBindEnv *, corTermLink*&, corBindEnv *, corBindEnv * ) 
{
 corCArg c(this);
 return c;
}

/*****************************************************************************/
/*
		Copy and simplify from old_arglist to new_arglist. 
   		Return the number of variables in new_args (return -1 if this is 
		unknown). 
*/

int corCopyArgs(register corArgList * new_arglist, register corArgList * old_arglist,
			 corBindEnv *env, int arity, corBindEnv *dont_rename_env,
			 corBindEnv *rename_only_env)
{

	ASSERT((rename_only_env==NULL));
	if (dont_rename_env == NULL)
		dont_rename_env = &corUnusedEnv;
	corTermLink *renamed_vars = NULL;
	register int i;
	for (i = 0; i < arity; i++) {
		(*new_arglist)[i] = ((*old_arglist)[i]).simplify(env, renamed_vars, 
						dont_rename_env, rename_only_env);
	}

	for (i = 0; renamed_vars != NULL; i++) {
		corTermLink *tlink = renamed_vars;
		renamed_vars = tlink->next;
		delete tlink;
	}

	int newsize = dont_rename_env->max_var_number() + i;
	if (dont_rename_env != &corUnusedEnv) {
		dont_rename_env->resize( newsize ) ;
	}
	return newsize;
}

/*****************************************************************************/
int corCopyArg(corCArg& new_arg, corCArg& old_arg, corBindEnv *env)
{
	corBindEnv *dont_rename_env = &corUnusedEnv;
	corTermLink *renamed_vars = NULL;
	register int i;
	new_arg = old_arg.simplify(env, renamed_vars, dont_rename_env, NULL);

	for (i = 0; renamed_vars != NULL; i++) {
		corTermLink *tlink = renamed_vars;
		renamed_vars = tlink->next;
		delete tlink;
	}

	int newsize = dont_rename_env->max_var_number() + i;
	return newsize;
}

/*****************************************************************************/
/*
		This function copies out old_arglist into new_arglist.  It dereferences
		variables when copying and if the variable is in rule_env, it is
		renamed to be in dont_rename_env.  renamed_vars is a list of 
		rule variables that were renamed earlier by a call to 
		backpatch_bindenvs.

		Oddities/Quirks ::

		ASSUMPTION:  backpatching has already been done, so that all bindings
		are either to dont_rename_env or to rule_env.

*/

int corSimpleCopyArgs(corArgList *new_arglist, corArgList *old_arglist,
				corBindEnv *rule_env, corBindEnv *dont_rename_env,
				corTermLink *&renamed_vars)
{
	register int i;

	int arity = old_arglist->arity();

	for (i = 0; i < arity; i++) {
	  (*new_arglist)[i] = (*old_arglist)[i].simplify(rule_env,renamed_vars, dont_rename_env, /*rename_only_env=*/ rule_env);
	}

	for (i = 0; renamed_vars != NULL; i++) {
		corTermLink *tlink = renamed_vars;
		renamed_vars = tlink->next;
		delete tlink;
	}

	if (dont_rename_env != &corUnusedEnv) {
		dont_rename_env->resize(dont_rename_env->max_var_number() + i ) ;
	}

	return dont_rename_env->max_var_number() + i;
}	 

/*****************************************************************************/
/*
		backpatch_bindenvs

	(Author:  Sudarshan)
	(This function is related to return-unification.)

	This function backpatches bindings by going through the trail, in
	order to pretend as if the rule variables were renamed before unification.

	Any variables that were bound to a rule variable are now re-bound
	after dereferencing them.  
	
	If the dereferencing leads to a 
	binding in a bindenv other than rule_env or dont_rename_env,
	backpatch_bindenvs fails.  This cannot happen when return-unification
	succeeds, since there is only one bindenv other than rule_env.
	Failure also cannot happen for the case of rules with exactly one 
	derived body literal.

	If the dereferencing leads to a structure interpreted in the rule
	bindenv, this structure is renamed, using calls to arg::simplify.
	Note that as before, if other bindenvs are encountered, failure results.
	If a structure is interpreted in dont_rename_env, it need not be
	recursively traversed --- before rule application, the bindenv has 
	only local bindings, and hence there are no cross bindings.  Any cross
	bindings are created during rule application, and are hence backpatched
	here.

	The new variable names start from after the last variable in 
	dont_rename_env.  
	
	A list of renamed rule variables is returned in 
	renamed_vars - the ordering corresponds to the new names of the 
	variables.

	This backpatching does not rename free variables in the head of the
	rule that are not pointed to by any variables that got bound during
	earlier unification.
	Renaming those variables is handled separately after the call to
	backpatch_bindenvs.

	Note that rule variable bindings are not changed during backpatching,
	since these bindings have to be undone on backtracking.
	Backpatching conflicts with backtracking -- if X was bound to Y, and 
	later Y was bound to 1, backpatching may bind X to 1.  When backtracking
	reaches Y, it is unbound, and X ought to be bound to Y at this point.
	But path-compression during backpatching will screw up, since X will
	stay bound to 1.

	(See note re. how fact bindenvs are handled in this respect.)

	Note that backpatch_bindenvs takes time proportional to the size of the
	rule and to the number of bindings that were created in trail in the
	course of this inference (multiplied by the acceess cost of variables
	in the bindenvs).

	return status:
		0	   - failure - can't backpatch the bindings.
		1	   - success

*/

int corbackpatch_bindenvs(corStackMark start, corBindEnv *dont_rename_env, corBindEnv *rule_env, corTermLink *&renamed_vars) 
 {
  corTerm binding;
  
  for (int i= start.trail_index(); i < corCurTrail->TrailSize; i++) {
    if ((*corCurTrail)[i].var < 0) // Not a variable binding
      continue;
    
    // WARNING:
    //  In the following, do not change rule var bindings --
    //  they have to be restored on backtracking,
    //  and backpatching can screw up backtracking
    //  the fact bindenv (there is only one) can be backpatched
    //  since it is copied before backpatching.
    // 
    //  This does not affect the complexity of the Opt-NG-SN
    //  algorithm, since  the amount of work done in backpatching
    //  each binding is quite small

    binding = (*corCurTrail)[i].bindenv->lookup((*corCurTrail)[i].var);

    if (binding.bindenv == dont_rename_env)
      continue;  // No need to fix the binding
    
    corfull_dereference_term(binding);

    if (binding.bindenv == dont_rename_env) {
      if ((*corCurTrail)[i].bindenv != rule_env)
	(*corCurTrail)[i].bindenv->bind((*corCurTrail)[i].var, binding);
      continue;
     }
    else if (binding.bindenv != rule_env) {
      if ( ! binding.expr.isConstant())
	// Binding to another bindenv -- cannot backpatch
	goto fail_return;

      // binding.expr is a constant.  Can use dont_rename_env for it
      binding.bindenv = dont_rename_env;
      if ((*corCurTrail)[i].bindenv != rule_env)
	(*corCurTrail)[i].bindenv->bind((*corCurTrail)[i].var, binding);
      continue;
     }
    else  { 		// binding.bindenv == rule_env
      corCArg deref_arg = binding.expr.simplify(binding.bindenv,
					     renamed_vars, dont_rename_env, rule_env);
      if (deref_arg.IsEmpty()) // patching of binding failed.
	goto fail_return;
      binding.expr = deref_arg;
      binding.bindenv = dont_rename_env;
      if ((*corCurTrail)[i].bindenv != rule_env)
	(*corCurTrail)[i].bindenv->bind((*corCurTrail)[i].var, binding);
      continue;
     }
   }

  return 1;

 fail_return:
  for (i = 0; renamed_vars != NULL; i++) {
    corTermLink *tlink = renamed_vars;
    renamed_vars = tlink->next;
    delete tlink;
   }
  return 0;
 }


/* not used -wX
ArgList* SimplifyVars(BindEnv *env, int num_vars, Symbol** rule_var_names )
{
	TermLink *renamed_vars = NULL;
	register int i;
	ArgList *new_args = new ArgList(num_vars);
	Name dummyvarname = EnterSymbol("Xerror");

// 	Variables in env should not be renamed:  These are the variables in 
// 		the query.  If we have ?q(X,Y), and bindings X->X0, X0->Y, 
// 		we would otherwise rename Y in the bindings to say X_0, but not be 
// 		able to change the variable Y in the query.  The result would then 
// 		read X=X__0, Y=Y, which is wrong 

	for (i = 0; i < num_vars; i++) {
		VarArg *var = new VarArg (rule_var_names[i], i);
		(*new_args)[i] = (Arg *) var->simplify(env, renamed_vars, env, NULL);
		if ( (*new_args)[i] != var) delete var;
	}
	for (i = 0; renamed_vars != NULL; i++) {
		TermLink *tlink = renamed_vars;
		renamed_vars = tlink->next;
		delete tlink;
	}
	return(new_args);
}
*/

/*****************************************************************************/
corVarLink * corArg::var_list(corVarLink *rest)
{
 return rest;
}

/*****************************************************************************/
Variable corArg::max_vars()
{
	corVarLink *rest, *prev, *cur;
	Variable max_var_num=-1;

	rest = this->var_list(NULL);
	for( cur=rest;  cur; ) {
		if (cur->arg->var > max_var_num)
			max_var_num = cur->arg->var;
		prev = cur;
		cur = cur->next;
		delete prev;
	}
	return max_var_num+1;

}

/*****************************************************************************/
int corArg::equals(corArg *)
{
	fprintf(corexEnv->error_file, 
	"System Error: equals() called on class for which it is not defined\n");
	return 0;
}

/*****************************************************************************/
HashVal corArg::hash(corBindEnv* /*env = 0*/) 
{
	// Default function - overridden in subclasses.
	return VarHashValue;
}

/*****************************************************************************/
int corArg::scan_vars(void *, VarFunc *)
{
	return 0;
}

/*****************************************************************************/

