/*****************************************************************************/
//    BuiltinUserTupleRelation.C
/*****************************************************************************/

#include "BuiltinUserTupleRelation.h"

#include "TupleIterator.h"

/*****************************************************************************/
corBuiltinUserTupleRelation::corBuiltinUserTupleRelation(int arity, Name n,
    UserTupleProc user_op)
    : corBuiltinRelation(arity, n, NULL)
{
    op = user_op;
}

/*****************************************************************************/
corBindEnv* corBuiltinUserTupleRelation::get_next(corTupleIterator& iterator)
{
    corTuple *tuple = get_next_tuple(iterator);
    if (iterator.no_match())
        return NULL;
    else
        return iterator.bindenv;
}

/*****************************************************************************/
corTuple* corBuiltinUserTupleRelation::get_next_tuple(corTupleIterator& iterator)
{
    corBindEnv *env = iterator.bindenv;

    if (iterator.arg_list.arity() != arity()) {
        iterator.set_no_match(); 
        return NULL;
    }

    if (iterator.ipos > 0) { 
        iterator.stack_mark.pop_to();
        iterator.set_no_match();
        return NULL;
    }
    iterator.ipos++;
    iterator.stack_mark.get_mark();

    // first create the query tuple from the bindenv
    corArgList *alist = new corArgList(arity());
    corBindEnv *queryenv = NULL;

    for (int i = 0; i < arity(); i++) {
        corTerm term;
        term.expr = iterator.arg_list[i];
        term.bindenv = env;
        //FULL_DEREFERENCE_TERM(term);
        cordeep_dereference_term(term);
        if (!(((term.kindof() == COR_VARIABLE) && (term.bindenv == env))
            || (term.isConstant())))
        {
            term.expr = iterator.arg_list[i];
                // Dereferencing lead to a
                // different bindenv -- cannot use dereferenced term.expr
                // with iterator.bindenv, so revert to original arg_list[i]
            queryenv = iterator.bindenv;            
        }
        (*alist)[i] = term.expr;
    }

    corTuple query(alist, queryenv); 


    // now call the user routine to compute the (single) answer to the query
    corTuple *result = (corTuple *) (*op)(&query); 

    if (result == NULL) {
        // no answer to query
        iterator.stack_mark.pop_to();
        iterator.set_no_match(); 
        return NULL;
    }

    if (corunify_arg_lists(iterator.arg_list, result->args(), env,
        result->bindenv))
    {
        iterator.reset_no_match();
        return result;
    }
    else {
        iterator.stack_mark.pop_to();
        iterator.set_no_match();
        return NULL;
    }
}

/*****************************************************************************/

