/*****************************************************************************/
//    Execute.C
/*****************************************************************************/

#include <stdio.h>

#include "Execute.h"

#include "ArgList.h"
#include "StorageRelation.h"
#include "BindEnv.h"
#include "ModuleInfo.h"
#include "SccInfo.h"
#include "Globals.h"
#include "HashSimpleRelation.h"
#include "Compile.h"

#include "Suspend.h"
#include "Suspension.h"

/*****************************************************************************/

#ifdef WITH_PERSISTENCE
extern corStorageRelation *make_pers_relation(Name, int arity, int delta_index,
                       int set_subsum, corArgList *a_list);
#define MAKE_PERS_REL_MACRO(a,corb,c,d,e) make_pers_relation(a,corb,c,d,e)
#else
#define MAKE_PERS_REL_MACRO(a,corb,c,d,e) cormake_local_relation(a,corb,c,d)
#endif

/*****************************************************************************/

corArgList *corMagicArgs = NULL;
corBindEnv *corMagicBindEnv = NULL;
corStorageRelation *corMagicRelation = NULL;

// points to current moduleInfo defaults
corModuleInfo *corCurrModuleInfo = NULL;

/*****************************************************************************/
void corPrintSccRecursiveRules(corSccInfo* scc, int start, FILE *outf)
{
    int j;

    fprintf(outf,"\n[Iterating over recursive rules ...]") ;
    fprintf(outf,"\n---------------------------------------\n");
     
    for (int i = start; i < scc->numSNRules ; i++) {
        fprintf(outf,"%s\n",scc->sn_rules[i].rInfo->pInfo.rule);
        fprintf(outf, "Predord :: ");
        for (j = 0; j < scc->sn_rules[i].rInfo->num_literals; j++)
            fprintf(outf, "%d ",scc->sn_rules[i].predOrders[j]);      
        fprintf(outf, "\n1st bt :: ");
        for (j = 0; j < scc->sn_rules[i].rInfo->num_literals; j++)
            fprintf(outf, "%d ",scc->sn_rules[i].firstFailureBacktrack[j]);
        fprintf(outf, "\nnext bt :: ");
        for (j = 0; j < scc->sn_rules[i].rInfo->num_literals; j++)
            fprintf(outf, "%d ",scc->sn_rules[i].nextFailureBacktrack[j]);
        fprintf(outf, "\n\n");
    }
    
    fprintf(outf,"---------------------------------------\n\n");
    fflush(outf);
}

/*****************************************************************************/
void corPrintSccNonRecursiveRules(corSccInfo* scc, FILE *outf)
{
    fprintf(outf,"\n[Applying non-recursive rules ...]") ;
    fprintf(outf,"\n---------------------------------------\n");
    
    for (int i = 0; i < scc->numSNRules ; i++) {
        if (scc->sn_rules[i].recursive) break ;
//        if (!scc->sn_rules[i].ignore_this_rule) {
            fprintf(outf,"%s\n",scc->sn_rules[i].rInfo->pInfo.rule);
            fprintf(outf, "\n");
//        }
    }
    
    fprintf(outf,"---------------------------------------\n\n");
}

/*****************************************************************************/
void corEndOfScc(corSccInfo* scc, int iters, FILE *outf)
{
    int i;

    if (corexEnv->profile_scc) {
        for (i=0; i< scc->numSNRules; i++) {
            //if (!scc->sn_rules[i].ignore_this_rule) {
                scc->sn_rules[i].rInfo->pInfo.print(outf, "SCC",
                    scc->sn_rules[i].predOrders,
                    scc->sn_rules[i].rInfo->num_literals);
            //}
        }
        fprintf(outf, "\nNumber of SCC iterations = %d\n", iters);
    }
}

/*****************************************************************************/
static void clearStatistics(corQFMaterializedModuleData *md, FILE *outf)
{
    int i, j;
    long profile_rules = ((corexEnv->GlobalRelationOptions | md->rel_options) &
        REL_PROFILE_STATS);
    
    if (profile_rules && (corexEnv->profile_module)) {
        md->module_info.corGlobalProfileInfo.clear();
        md->module_info.corGlobalProfileInfo.set_rule_name("");
        for (i = 0; i < md->numSccs; i++) {
            for (j=0; j < md->ruleArray->numRules; j++) {
                corAddProfileInfo(&(md->ruleArray->rInfoPtrs[j]->pInfo),
                    &(md->module_info.corGlobalProfileInfo));
                md->ruleArray->rInfoPtrs[j]->pInfo.clear();
            }
        }

        md->module_info.corGlobalProfileInfo.print(outf, "MODULE SUMMARY");
        md->module_info.corGlobalProfileInfo.clear();
    }
    else {
        for (i = 0; i < md->numSccs; i++) {
            for (j = 0; j < md->sccArray[i]->numSNRules; j++) {
                md->sccArray[i]->sn_rules[j].rInfo->pInfo.clear();
            }
        }
    }
}


/*****************************************************************************/
/*
    WARNING  - with recursion deeper than exEnv->C_max_recursion_depth_default
    this code will complain and abort.
*/

corStorageRelation *corfind_local_relation(Name name, int arity)
{
    if (corT_Stack.count <= 0 || corT_Stack.top() == NULL)
        return NULL;

    int offset = corget_offset(corT_Stack.top(), name, arity, 0);
    if (offset < 0)
        return NULL;

    return (corStorageRelation*)(corT_Stack.top()->rels[offset].relPtr);
}

/*****************************************************************************/
// Used by psn to determine whether to insert the query tuple before
// evaluating the scc, or during scc evaluation. the function in_scc
// returns non-zero if the Relation rel corresponds to one of the
// recursive predicates of the scc.

static int in_scc(corSccInfo *scc, corStorageRelation *rel)
{
    for (int i = 0; i < scc->numPreds; i++) {
        if ((scc->pred_ptrs[i]->name == rel->name) &&
            (scc->pred_ptrs[i]->arity == rel->arity()))
                return 1;
    }
    return 0;
}

/*****************************************************************************/
static void init_rule_pointers(corTable *F, corQFModuleData *md,
    corStorageRelation *magic, int reused_table, int old_magic_relation)
{
    corStorageRelation *temp, *temprel;
    corModulePred *predicates = md->predicates->next;
    F->moduledata = md;

    //    check whether it is a non SaveModule or the first invocation
    //    of a SaveModule
    int new_init = ((!md->module_info.SaveModule) ||
        (! ((corQFMaterializedModuleData *) md)->saved_state));
    
    for (; predicates; predicates = predicates->next) {
        Name name = predicates->pred->name;
        int arity = predicates->pred->arity;
        int offset = predicates->offset;
        switch (predicates->kind) {
        case ExternalPredKind: 
            //    If reused_table is set, we should be able to avoid the 
            //    overhead of finding external relations all over again.    
            //    However if anything has changed externally (say, because 
            //    of a rel_close), we need to do this.
            //    Perhaps we should maintain a flag and find relations again
            //    only if required
            F->rels[offset].AssignRel(corFind_External_Relation(name, arity));
            F->rels[offset].relPtr->local_options =
                md->rel_options | corexEnv->GlobalRelationOptions;
            break;

        case QueryPredKind:
            if (new_init) {
                //    the condition tested is not (! reused_table) as for 
                //    LocalPredKind below since answer relations are used 
                //    at the calling point even after the table is returned 
                //    to the free table list.
                if (predicates->disk_rel) {
                     temp = MAKE_PERS_REL_MACRO(name, arity,
                        md->module_info.IndexDeltas, predicates->check_subsum,
                        predicates->schema_def);
                     F->rels[offset].AssignRel(temp);
                     temp->ordsearch = md->module_info.UseOrdSearch;
                     if ((md->module_info.SaveModule) &&
                        (! ((corQFMaterializedModuleData *) md)->saved_state))
                    {
                        temp->getMark();
                    }
                }
                else {
                     temp = cormake_local_relation(name, arity,
                        md->module_info.IndexDeltas, predicates->check_subsum);
                     F->rels[offset].AssignRel(temp);
                     temp->ordsearch = md->module_info.UseOrdSearch;
                     if ((md->module_info.SaveModule) &&
                        (! ((corQFMaterializedModuleData *) md)->saved_state))
                    {
                         temp->getMark();
                    }
                }
                F->rels[offset].relPtr->local_options = md->rel_options |
                    corexEnv->GlobalRelationOptions;
            }
            break;

        case LocalPredKind:
            if (! reused_table) {
                //    Relation not already set up.  With SaveModules, the
                //    relation should also have all the tuples computed until
                //    the last time.
                 
                ASSERT(new_init != 0)
                
                if (predicates->disk_rel) {
                     temp = MAKE_PERS_REL_MACRO(name, arity,
                        md->module_info.IndexDeltas, 
                        predicates->check_subsum, predicates->schema_def);
                }
                else {
                     temp = cormake_local_relation(name, arity,
                        md->module_info.IndexDeltas, predicates->check_subsum);
                }
                
                F->rels[offset].AssignRel(temp);
                temp->ordsearch = md->module_info.UseOrdSearch;             
                if ((md->module_info.SaveModule) &&
                    (! ((corQFMaterializedModuleData *) md)->saved_state))
                {
                     temp->getMark();
                }
            }
            else {
                F->rels[offset].relPtr->local_options =
                    md->rel_options | corexEnv->GlobalRelationOptions;
                continue;
            }
            
            F->rels[offset].relPtr->local_options =
                md->rel_options | corexEnv->GlobalRelationOptions;
            break;
            
        case MagicQueryPredKind:
            if ((md->module_info.UseOrdSearch) && (! F->context)) {
                F->context = new corContext();
            }
            
            if (new_init) {
                if (predicates->disk_rel) {
                     F->rels[offset].AssignRel(MAKE_PERS_REL_MACRO(name, arity,
                        md->module_info.IndexDeltas, predicates->check_subsum,
                        predicates->schema_def));
                }
                else {
                     F->rels[offset].AssignRel(cormake_local_relation(name, arity,
                        md->module_info.IndexDeltas, predicates->check_subsum));
                }
                old_magic_relation = 0;
                if ((md->module_info.SaveModule) &&
                    (! ((corQFMaterializedModuleData *) md)->saved_state))
                {
                     ((corStorageRelation *)(F->rels[offset].relPtr))->getMark();
                }
            }
            // else retain the old relation 
                
            // NOTE: the tuples in "magic" have not been added as yet
            temprel = (corStorageRelation *)(F->rels[offset].relPtr);
            temprel->ordsearch = md->module_info.UseOrdSearch;
            temprel->r_kind = COR_R_MAGIC;
            
            if (md->module_info.UseOrdSearch) {
                if (! temprel->for_ordsearch) {
                     temprel->for_ordsearch = new corHashSimpleRelation(arity);
                }
                magic->add_all_goals(temprel->for_ordsearch, F->context,
                    temprel);             
            }

            temprel->local_options =
                md->rel_options | corexEnv->GlobalRelationOptions;

            break;
            //    From this point on, the passed in relation "magic"
            // is useless.    However, it gets reclaimed only in 
            // clear_rule_pointers.
            
        }
        
        struct corAnnoStruct *cur;
        
        for (cur= predicates->index_annos; cur; cur=cur->next) {
            if (cur->anno->arglist2) {
                F->rels[offset].relPtr->add_index(cur->anno->arglist1,
                    cur->num_vars, cur->anno->arglist2);
            }
            else {
                F->rels[offset].relPtr->add_index(cur->anno->arglist1);
            }
        }
        
        for (cur= predicates->aggsel_annos; cur; cur=cur->next)
            F->rels[offset].relPtr->add_agg_sel_anno(cur->anno);
        
        for (cur = predicates->prior_annos; cur; cur=cur->next)
            F->rels[offset].relPtr->add_prioritize_info(cur->anno);

        if (predicates->kind == MagicQueryPredKind) {
            if (!md->module_info.UseOrdSearch) {
                // first add the tuples in "magic" to the relation
                // there could be marks because of save modules
                magic->add_all_tuples((corStorageRelation *)(F->
                    rels[offset].relPtr), 1, 1);
                     
                if (((corQFMaterializedModuleData *)md)->use_psn) {
                     corMagicRelation = (corStorageRelation *)(F->rels[offset].relPtr);
                     if (!in_scc(((corQFMaterializedModuleData *)md)->sccArray[0],
                        corMagicRelation))
                    {
                        corMagicRelation->insert_new(*corMagicArgs, corMagicBindEnv);
                        corMagicRelation = NULL;
                        corMagicArgs    = NULL; corMagicBindEnv = NULL;
                    }
                }
            }
            else {
                // get the fact from the Context
                int any_new_fact_in_done = 0;
                corGoalNode *g_node =
                    F->context->get_unmarked_goal(any_new_fact_in_done);
                
                ASSERT((g_node != NULL));
                ASSERT((g_node->owner_rel == F->rels[offset].relPtr));
                if ((!md->module_info.SaveModule) ||
                    (! g_node->owner_rel->check_subsum) ||
                    (! g_node->owner_rel->is_subsumed(g_node)))
                {
                     g_node->owner_rel->insert_tuple(g_node);
                }
            }
        }
    }
}

/*****************************************************************************/
static int occurs_in_module(corModulePred *predicates, Name name)
{
    for (predicates = predicates->next; predicates;
        predicates = predicates->next)
    {
        if (name->equals(predicates->pred->name)) return 1;
    }
    return 0;
}

/*****************************************************************************/
static void clear_rule_pointers(corTable *F, corQFModuleData *md,
    corStorageRelation *magic)
{
    corModulePred *predicates = md->predicates->next;
    F->moduledata = md;

    for (; predicates; predicates = predicates->next) {
        Name name = predicates->pred->name;
        //int arity = predicates->pred->arity;
        int offset = predicates->offset;
        corStorageRelation *temprel = (corStorageRelation *)(F->rels[offset].relPtr);
        switch (predicates->kind) {
        case ExternalPredKind: 
            break;
        case QueryPredKind:
            if ((md->module_info.SaveModule) &&
                (! ((corQFMaterializedModuleData *) md)->saved_state))
            {
                temprel->free_allMarks();
            }
            if (md->module_info.SaveModule) {
                temprel->getMark();
            }
            break;
        case LocalPredKind:
        {
            if (temprel->for_ordsearch) {
                // this can be done even for SaveModules
                temprel->for_ordsearch->empty_relation(0);
            }
            if (! md->module_info.SaveModule) {
                if (temprel->mp_done_reln && !occurs_in_module(md->predicates,
                    temprel->mp_done_reln->name))
                {
                    // for OrdSearch
                    temprel->mp_done_reln->empty_relation(0);
                }
                
                // the mp_done relation does not have copies of tuples
                if (coris_mp_done(temprel->name))
                    temprel->empty_relation(0);
                else
                    correclaim_relation(temprel);                
            }
            else {
                // get appropriate mark after releasing previous marks
                temprel->free_allMarks();
                temprel->getMark();
                if (temprel->mp_done_reln && !occurs_in_module(md->predicates,
                    temprel->mp_done_reln->name))
                {
                    temprel->mp_done_reln->free_allMarks();
                    temprel->mp_done_reln->getMark();
                }
            }
            break;
        }
        case MagicQueryPredKind:
        {
            if (!md->module_info.SaveModule) {
                temprel->ordsearch = 0; 
                // else Ordered Search will crash the next time around
                if (temprel->mp_done_reln && !occurs_in_module(md->predicates,
                    temprel->mp_done_reln->name))
                {
                    temprel->mp_done_reln->empty_relation(0);
                }
            }
            else {
                // get appropriate mark after releasing previous marks
                temprel->free_allMarks();
                temprel->getMark();
                if (temprel->mp_done_reln && !occurs_in_module(md->predicates,
                    temprel->mp_done_reln->name))
                {
                    temprel->mp_done_reln->free_allMarks();
                    temprel->mp_done_reln->getMark();
                }
            }
             
            if (temprel->for_ordsearch) {
                // this can be done even for SaveModules
                temprel->for_ordsearch->empty_relation(0);
            }
             
            if (! md->module_info.SaveModule) {
                //    Change to reclaim some of the memory that is being wasted
                //    here temprel->empty_relation( /*deleteTuples =*/ 1 );
                correclaim_relation(temprel);
                F->rels[offset].relPtr = NULL;
            }
            break;
        }
        }
    }

    // reclaim "magic" 
    magic->empty_relation(0);
    magic->ordsearch=0;
    // else Ordered Search will crash the next time around

    if (md->module_info.SaveModule) {
        ((corQFMaterializedModuleData *) md)->saved_state = 1;
    }
}

/*****************************************************************************/
static corTable *initialize_do_query(corQFModuleData *md, corStorageRelation *magic,
    int old_magic_relation)
{
    // Make a copy of the table. This is to handle recursive calls to do_query.
    // Each call should see an unmodified Table.
    // Except in the case of SaveModule, where each call should see
    // the result of the last call to the module, and recursive calls
    // through a saved module are disallowed.
  
    corTable *table_copy;
    int reused_table = 0;

    if (md->module_info.SaveModule) {
        ASSERT((md->F != NULL));
        ASSERT((md->F->rels != NULL));
        // check whether this module was called for the first time 
        if (((corQFMaterializedModuleData *) md)->saved_state) {
             table_copy = md->F;
             reused_table = 1;
             // each call should see the result of the last call to the 
             // saved module.
        }
        else {
            table_copy = new corTable(md);
            delete md->F;
            md->F = table_copy;
        }
    }
    else {
        table_copy = new corTable(md) ;
    }

    init_rule_pointers(table_copy, md, magic, reused_table, old_magic_relation);
  
    return table_copy;
}

/*****************************************************************************/
void corcleanup_do_query(corQFModuleData *md, corStorageRelation *magic,
    corTable *table_copy)
{
    if (md->module_info.UseOrdSearch) {
        if (table_copy->context) {
            table_copy->context->free_all();
        }
    }
    clear_rule_pointers(table_copy, md, magic) ;

    // probably unnecessary here -- Praveen
    corMagicRelation = NULL;
    corMagicBindEnv = NULL;
    corMagicArgs = NULL;

    clearStatistics((corQFMaterializedModuleData *)md, corexEnv->trace_file);
}


/*****************************************************************************/
/*
    This function is called at run time to execute a query on the module.
    The magic fact for the query is in the magic relation magic.
    old_magic_relation specifies if the magic relation is new or old.
    The iterator is passed in to allow for a staggered bottom-up evaluation
    mode, where answers at the top level of the module are returned at the
    end of each iteration of the seminaive loop. This requires that the state
    of the seminaive computation has to be stored across calls to do_query,
    and the tuple_iterator is the only structure that can store this state
    information
*/

corStorageRelation *cordo_query(corQFModuleData *md, corStorageRelation *magic,
    int old_magic_relation, corTupleIterator *iterator)
{
    corTable *table_copy;
    corModuleInfo *temp_module_info;

#    ifdef DO_PROFILE
        md->module_info.corGlobalProfileInfo.clear();
#    endif

    // If staggered evaluation is not being performed, or this is the
    // first time do_query is being called on this module for this query
    // initialize the table of relations
    if (!(iterator) || !(iterator->state)) {
        // Make a copy of the table.  This is to handle recursive calls
        // to do_query.  Each call should see an unmodified Table.
        table_copy = initialize_do_query(md, magic, old_magic_relation);
    }
    else {
        table_copy = iterator->state->table;
    }

    corT_Stack.push(table_copy);
    // THIS SHOULD BE PART OF PUSH !!    PRAVEEN
    temp_module_info = corCurrModuleInfo;
    corCurrModuleInfo = &(table_copy->moduledata->module_info);

#    ifdef DEBUG
        if (corexEnv->dbg_modules) {
            fprintf(corexEnv->trace_file, "Entering module\n\n");
        }    
#    endif

    corStorageRelation *answer_rel = (corStorageRelation *)(table_copy->
        rels[corget_offset(table_copy, md->export_pred)].relPtr);

    // For the two cases, save_module + and eager_eval -, staggered evaluation
    // is not executed. Since md->Solve() is called with NULL as the iterator
    // argument, the module is completely evaluated, and the iterator->state
    // field remains NULL
    if (!iterator || md->module_info.SaveModule || !(md->module_info.EagerEval))
    {
        // no need to evaluate a no_magic module a second time around
        if ((! md->module_info.SaveModule) ||
            (! ((corQFMaterializedModuleData *) md)->no_magic) ||
            (!((corQFMaterializedModuleData *) md)->saved_state))
        {
            md->Solve(table_copy, NULL);
        }
        corcleanup_do_query(md, magic, table_copy);

        // adding this line to reclaim the magic relation allocated inside the
        // DerivedRel::get_next
        correclaim_relation(magic);


        corT_Stack.pop();

        // THIS SHOULD BE PART OF PUSH !!    PRAVEEN
        corCurrModuleInfo = temp_module_info;

        return answer_rel;
    }
    else {
        // if this is the first time the module is being called with the
        // iterator, the iterator->state field will be created.
        // otherwise, the existing iterator->state field will be used.
        // when there are no further answers, the state field is deleted and
        // set to NULL by SccInfo::Solve(), and this signifies that the
        // Table and temporary relations can be cleaned up.
        // this is the first invocation of the module using this iterator
        if (!(iterator->state)) {
            iterator->state = new corModuleExecInfo(md, table_copy, magic,
                cordo_query);
        }

        // this sets the rm_Arr field
        md->Solve(table_copy, iterator);

        // if no more answers
        if (!(iterator->state->rm_Arr)) {
            corcleanup_do_query(iterator->state->md, iterator->state->magic,
                iterator->state->table);
            // NOTE that the state->scan_mark field is redundant !!
            delete iterator->state;
            iterator->state = NULL;

        //  LEAK?  The top of stack is just thrown away,
        //  it should be freed.
            //Table *tos = T_Stack.top();
            //T_Stack.pop();
            //delete tos;

            corT_Stack.pop();

            corCurrModuleInfo = temp_module_info;
             // THIS SHOULD BE PART OF PUSH !!    PRAVEEN

            return answer_rel;
        }

        // even for lazy evaluation, pop the table stack, or else,
        // the table stack abstraction is lost !
        corT_Stack.pop();

        corCurrModuleInfo = temp_module_info;
        // THIS SHOULD BE PART OF PUSH !!    PRAVEEN

        return answer_rel;
    }
}

/*****************************************************************************/
//    Fill the Table with the appropriate relation pointers

static void fill_table(corTable *F)
{
    for (int i = 0; i < F->numRels; i++) {
        F->rels[i].AssignRel(corFind_External_Relation(F->rels[i].relName,
            F->rels[i].relArity));
        F->rels[i].relPtr->local_options = corexEnv->GlobalRelationOptions;
    }
}

//**************************************************************************
//    Executes a single rule from the parserStruct. This function is invoked
//    from the routines in ParseActions.C to execute imperative rules.
//**************************************************************************
int corexecute_single_rule(corParserStruct &corparserStruct)
{
    if (!corparserStruct.snrule) {
        corparserStruct.snrule = new corSemiNaive;
        
        // get the info out of the parserStruct, and create the RuleInfo
        // structure, relation Table and seminaive rule structure
        corcompile_single_rule(corparserStruct, corparserStruct.table,
            corparserStruct.rInfo_arr, *(corparserStruct.snrule));
     }
    
    // fill the table with the appropriate relation pointers
    fill_table(corparserStruct.table);
    
    // evaluate the rule
    if (coralConfig.server_mode) {    // automatic suspension

      corsuspension.set = corconvert_ans_2_set(corcurUser->num_answers);
    }

    int new_tuples = 
      corparserStruct.snrule->evaluate(corparserStruct.table,NULL, NULL, 1);
    
    return(new_tuples);
}

/*****************************************************************************/

