#ifndef _CRDBRELATION_H_
#define _CRDBRELATION_H_

#ifdef DO_RDB

// move this to generic-rel.h later...
#define COR_R_RDBREL (COR_R_OTHER + 50)

#include <assert.h>
// #include "Unify.h"
// #include "cor_error.h"
// #include "Globals.h"
#include "BuiltinRelation.h"
// #include "generic-rel.h"
// #include "interface.h"
#include "Rdb.h"

// This is a bit confusing:
// CRdbRelation is a part of the coral relation hierarchy,
// but RdbRelation is a part of the Rdb hierarchy.
// All CRdb... things are part of coral and talk to the rdb stuff

class corCRdbRelation
: public corRelation
{
  public:

    corCRdbRelation(const Name newName, const RdbJoin& newJoinInfo);

    ~corCRdbRelation();

    RdbRelation& RdbRel() const; // access the first RdbRelation
    
    //  override from Relation:
    
    virtual corBindEnv* get_next(corTupleIterator&); // defunct

    virtual corTuple* get_next_tuple(corTupleIterator& iterator); 
				// get the next tuple using the iterator
    
    virtual void release(corTupleIterator& iterator);
				// close the iterator
    
    virtual int insert_tuple(corTuple*); // inserts tuple in relation by 
				      // calling insert_new

    virtual int insert_new(corArgList& argList, corBindEnv* pBindEnv,
			   corBindEnv* = 0, corTuple* = 0, int = 0); 
				// check subsumption, then inserts 
				// into backend dbms.
				// the last three parameters are not used
    
    virtual tuple_delete(corTuple* tuple);	// delete all tuples matching 
					// the bindings of this tuple

    virtual void print_name(FILE* outf); // print the name of the 
					 // relation on outf

  private:
    
    const RdbJoin joinInfo;	// list of rdb relations to join
				// along with the join conditions

    RdbQueryVersions* queryVersions;

    RdbInsert* rdbInsert;

    RdbDeleteVersions* deleteVersions;
    

    RdbCursor* Init(corTupleIterator& iterator);
				// initialize the based upon the iterator.
				// create a new cursor if the iterator is at
				// the top.

    RdbCursor* Bind(corTupleIterator& iterator);
				// 

    corTuple* Done(corTupleIterator& iterator);
				// 
				// always returns NULL

    int AsRdbItemList(corArgList& argList, corBindEnv* pBindEnv,
		      RdbItemList& result, BitVector& adorn) const;
				// convert the argList (w/ pBindEnv)
				// to the result RdbItemList.
				// adorn[i] is set to true iff 
				// argList[i] is bound
				// returns: 0 on success
				//          1 on failure

    corArgList* AsArgList(RdbItemList& result);
				// convert RdbItemList to an argList
				// returns a "new" arglist which the caller
				// is responsible for freeing

    corCRdbRelation(const corCRdbRelation&); // not allowed

    corCRdbRelation& operator =(const corCRdbRelation&); // not allowed
};

/*****************************************************************************/

// These two classes are used to keep a list of coral variable names &
// their position in the arglist.  Then the list can be sorted by name
// and then you can easily find the joins between the relations.

class corCRdbVar {
  public:
    int rel;
    int col;
    Name name;			// variable name in arglist
    
    corCRdbVar(int newRel, int newCol, Name newName)
      { rel = newRel; col = newCol; name = newName; }
};

class corCRdbVarPtrList
: public DList<corCRdbVar>
{
  public:
    void SortByName();
};


/***************************************************************************/

class corFuncArg;			// forw. decl.
// Global functions:
int CreateCRdbRelation(Name newName, RdbJoin& join);
int CreateCRdbJoinRelation(corFuncArg* pfaNewRel, corBindEnv* pBindEnv);
void CreateCRdbJoinList(RdbJoinCondList& joinList, corCRdbVarPtrList& varList);

/***************************************************************************/

inline RdbRelation& corCRdbRelation::RdbRel() const
{
    return joinInfo.Rel(0); 
}

/***************************************************************************/

inline corBindEnv* corCRdbRelation::get_next(corTupleIterator&) // defunct
{
    assert(0);
    return NULL;
}

/***************************************************************************/
/***************************************************************************/

#endif // DO_RDB

#endif // _CRDBRELATION_H_
