/*****************************************************************************/
//    IncrLoader.C
/*****************************************************************************/

#ifdef INCRLOADER

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>

#include <fcntl.h>
#include <sys/exec.h>

    // Where is this supposed to be defined?
/*  #define    PAGSIZ 1024  */

#include "IncrLoader.h"
#include "Globals.h"
#include "TupleIterator.h"

/*****************************************************************************/

#define DUMMY_FN_NAME "_CORAL_dummy"
#define CORAL_FILENAME_TEMPFILE  "_fname_coral.temp"

extern char* coral_name;    // declared in main.C

/*****************************************************************************/

#ifndef __GNUC__
#ifndef UNIX_HP
extern "C"  char *sbrk(int);
extern "C"  read(int, char *, int);
#endif
extern "C"  close(int);
#endif

/*****************************************************************************/
// move this to config.h

#ifdef __GNUC__
#    define _CORAL_COMPILER   "gcc"
#    define _INCR_LOAD_OPTS   "-c -w"
#    define _CORAL_LOADER     "ld"
#else
#    define _CORAL_COMPILER   "CC"
#    define _CORAL_LOADER     "ld"
#    ifdef UNIX_MIPS
#        define _INCR_LOAD_OPTS   "-c -w -G 0"
#    else
#        define _INCR_LOAD_OPTS   "-c"
#    endif
#endif

/*****************************************************************************/

class exec_function {
public:
    caddr_t   func;
    exec_function *next;
    exec_function() { next = NULL; }
};

/*****************************************************************************/
static exec_function * incr_load(const char *fname)
{
    char filename[200];
    strcpy(filename, fname);

    char buf[1000] ;
    int fd ,len ;
    len = strlen(filename);

    // first determine the full pathname of this program !
    sprintf(buf, "which %s 2>/dev/null > %s", coral_name,
        CORAL_FILENAME_TEMPFILE);

    if (system(buf)) {
        fprintf(corexEnv->error_file,
            "unable to determine pathname of executing program :");
        fprintf(corexEnv->error_file, " assuming %s\n", coral_name);
    }
    else {
        FILE *fd;
        if (!(fd = fopen(CORAL_FILENAME_TEMPFILE, "r"))) {
            fprintf(corexEnv->error_file,
                "unable to determine name of executing program :");
            fprintf(corexEnv->error_file, " assuming %s\n", coral_name);
            unlink(CORAL_FILENAME_TEMPFILE);
        }
        else {
            fgets(coral_name, 200, fd);
            // the last character is a \n, and needs to be overwritten
            coral_name[strlen(coral_name)-1] = '\0';
            fclose(fd);
            unlink(CORAL_FILENAME_TEMPFILE);
        }
    }

    // Check that the file is a .o file
    if (filename[len-2] != '.') {
        fprintf(corexEnv->error_file,
            "Input file should be a .C file, a .S file\n");
#        ifdef UNIX_MIPS
            fprintf(corexEnv->error_file,
                "or a .o file compiled with the -G 0 option\n");
#        else
            fprintf(corexEnv->error_file, "or a .o file\n");
#        endif
        return NULL;
    }

    if ((filename[len-1] != 'o') || (filename[len-2] != '.')) {
        if ((filename[len-1] != 'C') && (filename[len-1] != 'S')) {
            fprintf(corexEnv->error_file,
                "Input file should be a .C file, a .S file\n");
#            ifdef UNIX_MIPS
                fprintf(corexEnv->error_file,
                    "or a .o file compiled with the -G 0 option\n");
#            else
                fprintf(corexEnv->error_file, "or a .o file\n");
#            endif
            return NULL;
        }
        if (filename[len-1] == 'S') {
            sprintf(buf, "translator -i %s -o %s", filename, filename);
            buf[strlen(buf)-1] = 'C';
            if (system(buf)) {
                fprintf(corexEnv->error_file,
                    "Could not run the translator on %s\n", filename);
                return NULL;
            }
            filename[len-1] = 'C' ;
        }    

        sprintf(buf, "%s %s -I%s/includes %s", _CORAL_COMPILER,
            _INCR_LOAD_OPTS, corexEnv->coral_path, filename);
        if (system(buf)) {
            fprintf(corexEnv->error_file, "Could not compile %s\n", filename);
            return NULL;
        }

#        ifdef DEBUG
            fprintf(corexEnv->error_file, "Compiled file %s\n", filename);
#        endif
        filename[len-1] = 'o';
    }

    FILE *fdes ;

    //    Create the loadable code with all references patched
    caddr_t oadx = (caddr_t)sbrk(0);
#    ifdef UNIX_MIPS
        caddr_t adx = ((char*)oadx) + NBPG - (((int)oadx) % NBPG);
#    else
        /** UNIX_SPARC **/
        caddr_t adx = ((char*)oadx) + PAGSIZ - (((int)oadx) % PAGSIZ);
#    endif

    sprintf(buf, "%s -N -T %X -A %s %s -o %s.new", _CORAL_LOADER, 
        adx, coral_name, filename, filename);
    if (system(buf)) {
        fprintf(corexEnv->error_file, "Failed to load\n");
        sprintf(buf, "%s.new", filename);
        unlink(buf);
        return NULL ;
    }

#    ifdef DEBUG
        fprintf(corexEnv->error_file,
            "created file %s.new with references resolved\n", filename);
#    endif

    sprintf(buf, "%s.new", filename);
    strcpy(filename, buf);

    //    Perform the incremental load
    caddr_t ldadx;

#    ifdef UNIX_MIPS
        fd = open(buf, O_RDONLY, 0);

        struct filehdr Filehdr;
        struct aouthdr Aouthdr;
        read(fd, (char *)&Filehdr, sizeof(struct filehdr));
        read(fd, (char *)&Aouthdr, sizeof(struct aouthdr));
        
        sbrk(NBPG - (((int)oadx) % NBPG));

        ldadx = (caddr_t)sbrk((int)(Aouthdr.tsize + 
                        Aouthdr.dsize + Aouthdr.bsize));


        lseek(fd, (int)(N_TXTOFF(Filehdr, Aouthdr)), SEEK_SET);
        read(fd, ldadx, ((int)(Aouthdr.tsize + Aouthdr.dsize + Aouthdr.bsize)));
#    endif UNIX_MIPS

#    ifdef UNIX_SPARC
        fd = open(buf, O_RDONLY);

        struct exec Exec;
        read(fd, (char *)&Exec, sizeof(struct exec));
        sbrk(PAGSIZ - (((int)oadx) % PAGSIZ));

        ldadx = (caddr_t)sbrk((int)(Exec.a_text + Exec.a_data + Exec.a_bss));
        read(fd, ldadx, ((int)(Exec.a_text + Exec.a_data)));
#    endif

    close(fd);

#    ifdef DEBUG
        fprintf(corexEnv->error_file, "load addr is %X \n", (int)ldadx);
#    endif

    // retrieve the relocated address of the extern function

    sprintf(buf, "nm %s | grep %s > _CORAL.load_tmp2", filename, DUMMY_FN_NAME);
    if (system(buf)) {
        fprintf(corexEnv->error_file,
            "Unable to locate installing function %s in relocated code\n",
            DUMMY_FN_NAME);
        unlink(filename);
        unlink("_CORAL.load_tmp2");
        delete ldadx;
        return NULL ;
    }

    if (!(fdes = fopen("_CORAL.load_tmp2", "r"))) {
        fprintf(corexEnv->error_file,
            "Unable to open temporary file _CORAL.load_tmp2\n");
        unlink(filename);
        unlink("_CORAL.load_tmp2");
        delete ldadx;
        return NULL;
    }

    // retrieve the address of the extern function
    int return_address;
    char temp_char;
    exec_function *func_list = new exec_function;

    if (fscanf(fdes, "%X %c %s", &return_address, &temp_char, buf) != 3) {
        fprintf(corexEnv->error_file,
            "Unable to recover relocated installing function address\n");
        fclose(fdes);
        unlink(filename);
        unlink("_CORAL.load_tmp2");
        delete ldadx;
        return func_list;
    }

    exec_function *temp; int count    = 0;
    temp = func_list->next;
    func_list->next = new exec_function;
    func_list->next->func = (caddr_t)return_address;
    func_list->next->next = temp;
    count++;
    
    while (fscanf(fdes, "%X T %s", &return_address, buf) == 2) {
        temp = func_list->next;
        func_list->next = new exec_function;
        func_list->next->func = (caddr_t)return_address;
        func_list->next->next = temp;
        count++;
    }

    fclose(fdes);
    unlink(filename);
    unlink("_CORAL.load_tmp2");

    return func_list;
}

/*****************************************************************************/
#define FAIL_RETURN {iterator.set_no_match(); return;}

void loadIncrFile(char *filename, corTupleIterator& iterator)
{
    typedef void (*PF)(...);
    
    exec_function *func_list = incr_load(filename);
    if (!func_list || !(func_list->next)) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- incremental load failure\n");
        delete [] filename;
        FAIL_RETURN;
    }
    while (func_list->next) {
        PF anewfunc = (PF) func_list->next->func;
        (*anewfunc)();
        exec_function *temp = func_list->next;
        delete func_list;
        func_list = temp;
    }
}

#undef FAIL_RETURN

/*****************************************************************************/

#endif INCRLOADER

