// **************************************************************************
//    Interface.C
// **************************************************************************

#include "Globals.h"
#include "ParseUtils.h"

#include "Interface.h"
#include "HashSimpleRelation.h"
#include "BuiltinUserRelRelation.h"
#include "BuiltinUserTupleRelation.h"

// **************************************************************************

#define NilArg corNilSymbol

// **************************************************************************
corArg *make_arg (char *name)
{
    return (corArg *) corEnterSymbol(name);
}

// **************************************************************************
corArg *make_arg (Name name)
{
    return (corArg *) name;
}

// **************************************************************************
corArg *make_var (int i)
{
    char buf[100];
    sprintf(buf, "V_%d", i);
    corArg *v = new corVarArg(corEnterSymbol(buf), i);

    return v;
}

// **************************************************************************
corArg *make_var (char *name, int i)
{
    corArg *v = new corVarArg(corEnterSymbol(name), i);
    return v;
}

// **************************************************************************
corArg *make_var (Name name, int i)
{
    corArg *v = new corVarArg(name,i);
    return v;
}

// **************************************************************************
corArg *make_arg (char *func_name, corArgList *args)
{
    corFuncArg *fa = corFuncArg::New(corEnterSymbol(func_name), args->count());

    for (int i=0; i < args->count(); i++) {
        fa->args[i] = (*args)[i];
        ASSERT((fa->args[i] != NULL));
    }
    return fa;
}

// **************************************************************************
corArg *make_cons (corArg *a, corArg *corb)
{
    corFuncArg *fa = corFuncArg::New(corConsSymbol, 2);

    fa->args[0] = a;
    fa->args[1] = corb;

    return fa;
}
corArg *make_cons (corCArg a, corCArg corb)
{
    corFuncArg *fa = corFuncArg::New(corConsSymbol, 2);

    fa->args[0] = a;
    fa->args[1] = corb;

    return fa;
}

// **************************************************************************
corArg *make_nil ()
{
    return ((corArg *) NilArg);
}

// **************************************************************************
// Return the length of a [...] type list, or < 0 if it is not a list.
//

int corlist_length( corCArg list )
{
    int cnt = 0;

    if (! is_list(list))
        return(-1);

    corFuncArg *fa = list.FuncArgP();
    
    while (! is_nil(fa)){

        cnt++;

        if (! is_list(fa))
            break;

        fa = fa->args[1].FuncArgP();
    }  

    return(cnt);
}

// **************************************************************************
// Make a list of the arguments given i.e (X,1,[A,B],M) -> [X,1,[A,B],M]
//

#define MAX_NUM_ARG   128

corArg *make_list(int num_arg, ...)
{
    int i, flg = 0;
    va_list ap;
    va_start(ap,num_arg);
   
    corArg **arg_ptr; 
    corArg *arg_array[MAX_NUM_ARG]; 

    corArg *list = make_nil();

    if (num_arg <= MAX_NUM_ARG)
        arg_ptr = arg_array;
    else {
        arg_ptr = new (corArg*)[num_arg];
        flg = 1;
    }

    for (i=0; i < num_arg; i++)
        arg_ptr[i] = va_arg(ap, corArg *);

    for (i=num_arg-1; i >= 0; i--) {
        list = make_cons(arg_ptr[i],list); 
    }

    if (flg)
        delete [] arg_ptr;
     
    va_end(ap);
    return(list);
}
#undef MAX_NUM_ARG

// **************************************************************************
corArgList *make_arglist (int len, ...)
{
    int i=0;
    va_list ap;
    va_start(ap,len);

    corArgList *al = new corArgList(len);

    for(i =0; i < len; i++) {
        //(*al)[i] = ( (Arg **) ap)[i];
        (*al)[i] = va_arg(ap, corArg *);
        ASSERT(((*al)[i] != NULL));
    }
    va_end(ap);
    return al;
}

// **************************************************************************
corArgList *make_arglist1 (corArg* a1)
{
    corArgList *alist = new corArgList(1);
    (*alist)[0] = a1;
    return alist;
}

corArgList *make_arglist1 (corCArg a1)
{
    corArgList *alist = new corArgList(1);
    (*alist)[0] = a1;
    return alist;
}

// **************************************************************************
corArgList *make_arglist2 (corArg* a1, corArg *a2)
{
    corArgList *alist = new corArgList(2);
    (*alist)[0] = a1;
    (*alist)[1] = a2;
    return alist;
}
corArgList *make_arglist2 (corCArg a1, corCArg a2)
{
    corArgList *alist = new corArgList(2);
    (*alist)[0] = a1;
    (*alist)[1] = a2;
    return alist;
}

// **************************************************************************
corArgList *make_arglist3 (corCArg a1, corCArg a2, corCArg a3)
{
    corArgList *alist = new corArgList(3);
    (*alist)[0] = a1;
    (*alist)[1] = a2;
    (*alist)[2] = a3;
    return alist;
}


corArgList *make_arglist3 (corArg* a1, corArg *a2, corArg *a3)
{
    corArgList *alist = new corArgList(3);
    (*alist)[0] = a1;
    (*alist)[1] = a2;
    (*alist)[2] = a3;
    return alist;
}

// **************************************************************************
corArgList *make_arglist4 (corArg* a1, corArg *a2, corArg *a3, corArg*a4)
{
    corArgList *alist = new corArgList(4);
    (*alist)[0] = a1;
    (*alist)[1] = a2;
    (*alist)[2] = a3;
    (*alist)[3] = a4;
    return alist;
}

// **************************************************************************
corArgList *concat_arglists (corArgList *al1, corArgList *al2)
{
    int c1 = al1->count();
    int c2 = al2->count();
    corArgList *alist = new corArgList(c1+c2);

    for (int i=0; i < c1; i++){
        (*alist)[i] = (*al1)[i];
    }

    for (int j=0; j < c2; j++){
        (*alist)[c1+j] = (*al2)[j];
    }
    return alist;
}

// **************************************************************************
corArgList *cormake_vararglist (int n)
{
    corArgList *alist = new corArgList(n);
    for (int i=0; i<n; i++)
        (*alist)[i] = make_var(i);
    return(alist);
}

// **************************************************************************
int is_int (corArg *a)
{
    return (is_num(a) && (((corNumArg *) a)->num_kindof() == COR_INTEGER));
}

// **************************************************************************
int is_short (corArg *a)
{
    return (is_num(a) && (((corNumArg *) a)->num_kindof() == COR_INTEGER));
}

// **************************************************************************
int is_long (corArg *a)
{
    return (is_num(a) && (((corNumArg *) a)->num_kindof() == COR_INTEGER));
}

// **************************************************************************
int is_float (corArg *a)
{
    return (is_num(a) && (((corNumArg *) a)->num_kindof() == COR_DOUBLE));
}

// **************************************************************************
int is_double (corArg *a)
{
    return (is_num(a) && (((corNumArg *) a)->num_kindof() == COR_DOUBLE));
}

// **************************************************************************
int is_string (corArg *a)
{
    return (a->kindof() == COR_SYMBOL_CONST);
}

// **************************************************************************
int is_num (corArg *a)
{
    return (a->kindof() == COR_NUM_CONST);
}

// **************************************************************************
int is_constant (corArg *a)
{
    // assuming only two types of constants
    return (is_string(a) || is_num(a));
}


// **************************************************************************
  int is_string (corCArg a)
{
    return (a.kindof() == COR_SYMBOL_CONST);
}

// **************************************************************************
  int is_num (corCArg a)
{
    return (a.kindof() == COR_NUM_CONST);
}

// **************************************************************************
  int is_constant (corCArg a)
{
    // assuming only two types of constants
    return (is_string(a) || is_num(a));
}

// **************************************************************************

int is_var (corCArg a)
{
    return (a.kindof() == COR_VARIABLE);
}

// **************************************************************************

int is_functor (corCArg a)
{
    return (a.kindof() == COR_FUNCTOR);
}

// **************************************************************************
int is_list (corCArg a)
{
    return ((is_functor(a)) && (( a.FuncArgP())->functor() == corConsSymbol));
}

// **************************************************************************
int is_nil (corCArg a)
{
    // What should this be? (other??)
    return (a.other == NilArg);
}

// **************************************************************************
int make_int(corCArg a)
{ 
    return a.NumIval(); 
}

// **************************************************************************
short make_short(corCArg a)
{ 
    return (short)a.NumIval();
}

// **************************************************************************
long make_long(corCArg a)
{
    return (long)a.NumIval();
}

// **************************************************************************
float make_float(corCArg a)
{
    return a.NumDval();
}

// **************************************************************************
  double make_double(corCArg a)
{
    return a.NumDval();
}

// **************************************************************************
char *make_string(corCArg a)
{
    char *result = new char[(a.SymbolP())->length() + 1];
    memcpy(result, a.SymbolP()->string(), (int)(a.SymbolP())->length()+1);
    return result;
}

// **************************************************************************
char *functor_name (corArg *a)
{
    corSymbol *f_n = ((corFuncArg *) a)->functor();
    char *result = new char[f_n->length() + 1];
    memcpy(result, f_n->string(), (int) f_n->length() +1);
    return result;
}
char *functor_name (corCArg a)
{
    corSymbol *f_n = ( a.FuncArgP())->functor();
    char *result = new char[f_n->length() + 1];
    memcpy(result, f_n->string(), (int) f_n->length() +1);
    return result;
}

// **************************************************************************
corArgList *functor_args (corArg *a)
{
    return (& ((corFuncArg *) a)->args);
}

corArgList *functor_args (corCArg a)
{
    return (& ( a.FuncArgP())->args);
}

// **************************************************************************
corArg *make_car (corCArg a)
{
    return ( a.FuncArgP())->args[0].other; 
}

corArg *make_car (corArg *a)
{
    return ((corFuncArg*) a)->args[0].other; 
}

// **************************************************************************
corArg *make_cdr (corArg *a)
{
    return ((corFuncArg *) a)->args[1].other;
}

corArg *make_cdr (corCArg a)
{
    return (a.FuncArgP())->args[1].other;
}

// **************************************************************************
corTuple *make_tuple(int arity)
{
    return new corTuple(arity);
}

// **************************************************************************
corTuple *make_tuple(corArgList *arglist)
{
    return new corTuple(arglist);
}

// **************************************************************************
corRelation *make_rel(char *rel_name, int arity)
{
    return cormake_relation(corEnterSymbol(rel_name), arity);
}

// **************************************************************************
void delete_id_tuple (corRelation *rel, corTuple *tuple)
{
   /* We just mark the tuple as deleted. The actual deletion from
    * the relation and indexes is done in a lazy fasion, when the tuple
    * is encountered in the course of a scan.
    */
    if (tuple->is_deleted()) {
        fprintf(corexEnv->error_file,
             "Error: delete_tuple: tuple previously deleted\n");
    }
    else {
        tuple->do_delete();
        (rel->count)--;

#        ifdef DO_TRACE
        if (corexEnv->GlobalRelationOptions & REL_DISPLAY_INSERTIONS) {
            fprintf(corexEnv->trace_file, "[Deleting %s(", corSymbolString(rel->name));
            tuple->args().print(tuple->bindenv,corexEnv->trace_file);
            fprintf(corexEnv->trace_file, ")]\n");
        }
#        endif
    }
}

// **************************************************************************

extern int corsubsumes_arg_lists(corArgList& left, corArgList& right,
    corBindEnv *left_bindings, corBindEnv *right_bindings);

// **************************************************************************
//    This is a value based delete, where every tuple subsumed by tuple is
//    deleted from the relation. 

void delete_tuple (corRelation *rel, corTuple *tuple)
{
    corTupleIterator it(rel, tuple);
    corStackMark stack_mark;
    int check_subsum = rel->check_subsum;

    for (;;) {
        corTuple *new_tuple = it.get_next_tuple();
        if (it.no_match())
            break;
        if (corsubsumes_arg_lists(tuple->args(), new_tuple->args(),
            tuple->bindenv, it.tuple_env))
        {
            delete_id_tuple(rel, new_tuple);
            if (check_subsum==0)
                break;
        }
        stack_mark.pop_to();
    }
    stack_mark.pop_to();
}

// **************************************************************************
//    update_tuple does not update indices yet.  Fix this bug sometime.

void update_tuple (corTuple *tuple, int arg_pos, corArg *new_val)
{
    if ((arg_pos < 0) || (arg_pos >= tuple->args().count())) {
        fprintf(corexEnv->error_file,
                 "Error: update_tuple, argument position %d ", arg_pos);
        fprintf(corexEnv->error_file,
                 "does not exist.\n");
        return /* 0 */ ;
    }
    else { // the argument position exists
        (tuple->args())[arg_pos] = new_val;
        return /* 1 */ ;
    }
}

// **************************************************************************
void update_tuple (corRelation *rel, corTuple *old_tuple, corTuple *new_tuple)
{
    delete_tuple(rel, old_tuple);
    rel->insert_new(new_tuple);
}

// **************************************************************************
int call_coral(char *exp_pred_name, corTuple *query, corRelation *&result)
{
    Name pred = corEnterSymbol(exp_pred_name);
    corRelation *rel = corfind_external_relation(pred, query->args().count());
    if (!rel) {
        fprintf(corexEnv->error_file,
            "Error: call_coral: Relation %s not found\n", exp_pred_name);
        return -1;
    }

    if (!result)
        result = new corHashSimpleRelation(rel->arity());
    else {
        if (rel->arity() != result->arity()) {
            fprintf(corexEnv->error_file,
     "Error: call_coral: arity mismatch between query and result relation\n");
            return -1;
        }
    }
        

    corArrayBindEnv env(query->args().max_vars());

    corTupleIterator it(rel, query->args(), &env, NULL, NULL);
    int answer_count = 0;
#    ifdef DO_TRAIL
        corStackMark stack_mark;
#    endif
    for (;;) {
        it.get_next_tuple();
        if (it.no_match())
            break;
        else {
            result->insert_new(query->args(), &env);
            answer_count ++;
        }
#        ifdef DO_TRAIL
            stack_mark.pop_to();
#        endif
    }
    return answer_count;
}

// **************************************************************************
/* this function is probably outdated --- PRAVEEN */

int call_coral(char *exp_pred_name, corRelation *queries, corRelation *&result)
{
    Name pred = corEnterSymbol(exp_pred_name);
    corRelation *rel = corfind_external_relation(pred, queries->arity());
    if (!rel) {
        fprintf(corexEnv->error_file,
                "Error: call_coral: Relation %s not found\n", exp_pred_name);
        return -1;
    }

    if (!result)
        result = new corHashSimpleRelation(rel->arity());
    else {
        if (rel->arity() != result->arity()) {
            fprintf(corexEnv->error_file,
     "Error: call_coral: arity mismatch between query and result relation\n");
            return -1;
        }
    }

    int answer_count = 0;

    corArrayBindEnv env0(queries->arity());
    corArgList *arglist0 = cormake_vararglist(queries->arity());
    // create an arglist with vars 0 through (arity-1) for it.

    corTupleIterator it0(queries,*arglist0, &env0, NULL, NULL);
            /** to iterate over tuples in queries **/

#    ifdef DO_TRAIL
        corStackMark stack_mark0;
#    endif
    for (;;) {
        it0.get_next();
        if (it0.no_match())
            break;
#        ifdef DO_TRAIL
            corStackMark stack_mark1;
#        endif
        corTupleIterator it1(rel, *arglist0, &env0, NULL, NULL);

        for(;;) {
            it1.get_next();
            if (it1.no_match()) break;
            else {
                result->insert_new(*arglist0, &env0);
                answer_count ++;
            }
#            ifdef DO_TRAIL
                stack_mark1.pop_to();
#            endif
        }
#        ifdef DO_TRAIL
            stack_mark0.pop_to();
#        endif
    }
#    ifdef DO_TRAIL
        stack_mark0.pop_to();
#    endif
    return answer_count;
}

// **************************************************************************
//    add_builtin is only used from builtin-rel.C. It is NOT part of the
//    C++ interface

int add_builtin(char *pred, int arity, UserRelProc pred_solver)
{
    corBuiltinUserRelRelation *dummy = new corBuiltinUserRelRelation (arity, 
        corEnterSymbol(pred), pred_solver);
    return 1; // Success
}

// **************************************************************************
int add_builtin (char *pred, int arity, UserTupleProc pred_solver)
{
    corBuiltinUserTupleRelation *dummy =
        new corBuiltinUserTupleRelation(arity, corEnterSymbol(pred), pred_solver);
    return 1; // Success
}

// **************************************************************************
static int fill_in_params(corParserStruct *pstruct, corParameterStruct *params,
    int& index)
{
    if (!params) return 0;

    struct corparameter_link *temp_param = (pstruct->parameters)->last;
    corCArg temp_arg;
    
    while (temp_param) {
        switch (params[index].kind) {
        case COR_INT:
            temp_arg = *((int *)(params[index].value_ptr));
            break;

        case COR_SHORT:
            temp_arg = *((short *)(params[index].value_ptr));
            break;

        case COR_LONG:		// long = 4 bytes 
            temp_arg = int( *((long *)(params[index].value_ptr)));
            break;

        case COR_FLOAT:
	    temp_arg = double (*((float *)(params[index].value_ptr)));
            break;

        case COR_DOUB:
            temp_arg = *((double *)(params[index].value_ptr));
            break;

        case COR_STR:
            temp_arg = make_arg(*((char **)(params[index].value_ptr)));
            break;
            
        default:
            fprintf(corexEnv->error_file,
                 "CORAL::error : erroneous parameter type\n");
            return 1;
        }
        *(temp_param->parameter_slot) = temp_arg; // plug into the ParserStruct!
        temp_param = temp_param->prev;
        index++;
    }
    return 0;
}

// **************************************************************************
// execute the i'th ParserStruct in the ExecutionList

int corExecutionList::execute(int i, int& param_index)
{
    if (i >= actual_no_of_terms)
        return 1;
    

    if ((param_index < no_of_params) &&
        fill_in_params(exec_terms[i], parameters, param_index))
    {
        fprintf(corexEnv->error_file,
                 "fill_in_params fails !\n");
        return 1;
    }

    ASSERT((param_index <= no_of_params));
    exec_terms[i]->execute();

    return 0;    
}

// **************************************************************************
int corExecutionList::execute()
{
    int param_index = 0;

    for (int i = 0; (i < actual_no_of_terms); i++) {
        execute(i, param_index);
    }    

    return 0;
}

// **************************************************************************
void corExecutionList::add(corParserStruct *pstruct)
{
    if (actual_no_of_terms >= max_no_of_terms) {
        corParserStruct **temp = exec_terms;
        max_no_of_terms = max_no_of_terms + step_size;
        exec_terms = new corParserStruct*[max_no_of_terms];
        for (int i = 0; i < actual_no_of_terms; i++)
            exec_terms[i] = temp[i];        
    }
    exec_terms[actual_no_of_terms++] = pstruct;
}

// **************************************************************************
void corcopy_structs(corParserStruct* pstruct, corParserStruct& corparserStruct)
{
    *pstruct = corparserStruct;
    pstruct->rules->lastptr = &(pstruct->rules->chain);
    pstruct->rule_list = (corClause*)(pstruct->rules->chain);    

    corcopy_array(pstruct->rule_var_names, corparserStruct.rule_var_names);
    pstruct->sip_var_names.copy(corparserStruct.sip_var_names);
    corcopy_array(pstruct->rhspreds, corparserStruct.rhspreds);
}

// **************************************************************************
// process the top item on the input stack

void corprocessInputStack()
{
    corexEnv->C_exec_mode = COR_DOING_INTERPRET;

    if (! corparseStack->is_empty())
        corparseStack->parse(corexEnv->error_file,1);

    corexEnv->C_exec_mode = COR_DOING_IMPERATIVE;
}

// **************************************************************************
C_ScanDesc::C_ScanDesc(corRelation *rel)
{
    stackmark = NULL;
    env = new corArrayBindEnv(rel->arity());
    arg_list = cormake_vararglist(rel->arity());
    it = new corTupleIterator(rel, *arg_list, env);
}

// **************************************************************************
C_ScanDesc::C_ScanDesc(corRelation *rel, corTuple *match_tuple)
{
    if (!rel)
        return;
    stackmark = NULL;
    env = new corArrayBindEnv(rel->arity());
    arg_list = NULL;
    it = new corTupleIterator(rel, match_tuple->args(), env);
}

// **************************************************************************
corTuple *C_ScanDesc::next_tuple()
{

    if (stackmark) {
        stackmark->pop_to();
        /***  IS THIS NECESSARY ?? 
         CHNG:WX
           ***/
            delete env;
            env = new corArrayBindEnv(it->relation->arity());
        
    }

    corTuple *ret_tuple = it->get_next_tuple();

    // Note :: pipelining does not return an actual tuple. instead it merely
    //    provides bindings for variables. Since the interface expects
    //    a handle on a real tuple, we have to materialize it
    if (!it->no_match() && !ret_tuple) {
        for (int i = 0; i < it->arg_list.arity(); i++) {
            corTermLink *rv = NULL;
            it->arg_list[i].simplify(it->bindenv, rv, it->bindenv, NULL);
        }        
        ret_tuple = new corTuple(it->arg_list);
    }
    return ret_tuple;
}

// **************************************************************************
C_ScanDesc::~C_ScanDesc()
{
    delete stackmark;
    if (arg_list) delete arg_list;
    /***
        delete env;
    ***/
    delete it;
}

// **************************************************************************

