//**************************************************************************
//    Util.C
//**************************************************************************

#include <unistd.h>
#include <stdlib.h>

#include "Util.h"

#include "Assert.h"
#include "CoralError.h"
#include "Globals.h"

//**************************************************************************
FILE *coropen_file(char *name)
{
    FILE *tmp_file;
    if (name)
        tmp_file = fopen(name, "w");
    if (!tmp_file) 
    fprintf(corexEnv->error_file, "CORAL: unable to open file %s\n", name);
    return tmp_file;
}

//**************************************************************************
//    Reads in a specified file and dumps it on exEnv->error_file.  If there are
//    any problems opening the file, 1 is returned, else 0 is returned.
//**************************************************************************
int corDisplayFile(char *file_name)
{
    char file_buffer[1000] ;
    char buff[1000];
      
    ASSERT((strlen(file_name)<1000));
        if (coralConfig.server_mode){
              // Read the file and dump it into the socket
            FILE *fp = fopen(file_name,"r");
            if (fp == NULL)
              return(1);
            while(fgets(buff, 1000, fp) != NULL)
                fputs(buff,corexEnv->error_file); 
            fclose(fp);
            return(0);
        } else {
            sprintf(file_buffer, "more %s", file_name);
        }
    return system(file_buffer);
}

//**************************************************************************
int corconsultFile(char *filename, char *stack_filename)
{
    FILE *tmp_fp;

    tmp_fp = fopen(filename,"r");

    if (tmp_fp == NULL)
        return(0);
    else {
        corparseStack->handle_new_file(tmp_fp, stack_filename);
        if (stack_filename) {
            char temp_str[200];
            // Remove the <filename>.M file
            sprintf(temp_str, "%s.M", stack_filename);
            unlink(temp_str);
        }

            /* Can safely do this in server mode? */
        if (coralConfig.server_mode)
            corparseStack->parse(corexEnv->print_file,1);
    }
    return(1);
}

//**************************************************************************
int corconsultFile(char *filename, char *stack_filename, int coral_fd)
{
    FILE* tmp_fp;

    if( !( tmp_fp = fdopen(coral_fd, "r")))
        fprintf(corexEnv->error_file,
        "\n Could not open input file descriptor as a stream!!\n");

    if (tmp_fp == NULL)
        return(0);
    else {
        corparseStack->handle_new_file(tmp_fp, stack_filename);
        if (stack_filename) {
            char temp_str[200];
            // Remove the <filename>.M file
            sprintf(temp_str, "%s.M", stack_filename);
            unlink(temp_str);
        }
            /* Can safely do this in server mode? */
        if (coralConfig.server_mode)
            corparseStack->parse(corexEnv->print_file,1);
    }
    return(1);
}

//**************************************************************************
int corconsultBuff(char *buffer, char *buffname)
{
    FILE *tmp_fp;

    if (!buffer || (! *buffer))
        return(0);
    else {
        if (buffname) {
            char temp_str[200];
            // Remove the <buffname>.M file
            sprintf(temp_str, "%s.M", buffname);
            unlink(temp_str);
        }
        corparseStack->handle_new_buff(buffer,buffname);
            /* Can safely do this in server mode? */
        if (coralConfig.server_mode)
            corparseStack->parse(corexEnv->print_file,1);
    }
    return(1);
}

//**************************************************************************
char *cormk_tmp_file_name(char *name)
{
    static filecount = 0;
    sprintf(name, "/tmp/__coral_file_%d", filecount++);
    return (name);
}

//**************************************************************************
char *cormake_done_prefix()
{
    return "done_";
}

//**************************************************************************
//    Convert an adornment of the form "bbf" to a BitVector 110.  Return 0 if
//    ok, -1 otherwise.
//**************************************************************************
int corConvertAdornment(corBitVector& bits, char *adorn_string, int adorn_length)
{
    for (int i = 0; i < adorn_length; i++) {
        char adorn_char = *adorn_string++;
        switch (adorn_char) {
        case 'b':
            bits.setlen(i+1);
            bits.set(i, 1);
            break;
        case 'f':
            bits.setlen(i+1);
            bits.set(i, 0);
            break;
        default:
            return -1;
        }
    }
    return 0;
}

//**************************************************************************
int corStrHash(const char *s, int n)
{
    // simpler hash function : Praveen
    int hash = 0;
    for (n -= 1; n >= 0; n--) {
        hash += *s++;
    }
    return hash;
}

//**************************************************************************
int corhash_string(const char *text, int len)
{
    return corStrHash(text, len);
}

//**************************************************************************
