//**************************************************************************
//    ParseActions.C
//**************************************************************************

#include <stdio.h>

#include "ParseActions.h"
#include "ParserStruct.h"
#include "ParseUtils.h"
#include "BuiltinSymbols.h"
#include "Globals.h"
#include "Execute.h"
#include "Literal.h"
#include "Database.h"
#include "Interface.h"
#include "Tuple.h"
#include "VersionedBindEnv.h"
#include "Rewriting.h"
#include "Compile.h"

#include "Unparse.h"

//**************************************************************************

#define TMP_RELNAME  "__RelTemp"

//**************************************************************************

static int is_recursive(corParserStruct &corparserStruct);

//**************************************************************************
void corAppendFunction(corParserStruct& corparserStruct)
{
    corStackMark stack_mark;

    if (corexEnv->in_query_loop) {
        if (is_recursive(corparserStruct)) {
            fprintf(corexEnv->error_file,"CORAL::error :");
            fprintf(corexEnv->error_file,
                "Recursion not allowed in imperative rule!\n");
        }
        else {
            corcomplete_rule(corparserStruct.cur_rule, corparserStruct);
            int num_ans = corexecute_single_rule(corparserStruct);
            if (corexEnv->C_verbose_answer_default) {
                fprintf(corexEnv->print_file,
                    "(Number of Tuples to Append = %d)\n", num_ans);
            }
        }
    }
    else {
        fprintf(corexEnv->error_file, "CORAL::error :");
        fprintf(corexEnv->error_file, "Imperative statement inside module!\n");
    }
    if (corexEnv->C_verbose_answer_default) fflush(corexEnv->print_file);
    stack_mark.pop_to();   
}

//**************************************************************************
void corDeleteFunction(corParserStruct& corparserStruct)
{
    corStackMark stack_mark;

    if (corexEnv->in_query_loop) {
        if (is_recursive(corparserStruct)) {
            fprintf(corexEnv->error_file, "CORAL::error :");
            fprintf(corexEnv->error_file,
                "Recursion not allowed in imperative rule!\n");
        }
        else {
            corcomplete_rule(corparserStruct.cur_rule, corparserStruct);

            Name headName = corparserStruct.cur_rule->head->pred;
            Name tmpName  = corEnterSymbol(TMP_RELNAME);

            // rename the <head relation> to TMP_NAME
            if (!correnameRelation(headName, tmpName)) {
                fprintf(corexEnv->error_file, "unable to rename file\n");
                return;
            }

            // the tuples to delete are inserted into a new <head relation>
            int num_ans = corexecute_single_rule(corparserStruct);

            corRelation *rel = corfind_external_relation(headName,
                    corparserStruct.cur_rule->head->arity());
            
            corRelation *temprel = corfind_external_relation(tmpName,
                    corparserStruct.cur_rule->head->arity());

            if (!rel || !temprel) {
                fprintf(corexEnv->error_file, "Strange brew !!\n");
                return;
            }

            // Then delete the tuples in <head relation> from TMP_NAME
            corTuple *tuple ;
            FOR_EACH_TUPLE(tuple, rel) {
                temprel->tuple_delete(tuple);
            } END_EACH_TUPLE(tuple);

            // Finally, close the <head relation>
            rel->empty_relation(0);
            corDeleteInTab(corCurDB->RelationTable, headName);
            delete rel ;

            // rename TMP_NAME to <head relation>
            if (!correnameRelation(tmpName, headName)) {
                fprintf(corexEnv->error_file, "unable to rename tmp file\n");
                return;
            }

            if (corexEnv->C_verbose_answer_default) {
                fprintf(corexEnv->print_file,
                    "(Number of Tuples to Delete = %d)\n", num_ans);
            }
         }
    }
    else {
        fprintf(corexEnv->error_file, "CORAL::error :");
        fprintf(corexEnv->error_file, "Imperative statement inside module!\n");
    }
    if (corexEnv->C_verbose_answer_default) 
        fflush(corexEnv->print_file);
    stack_mark.pop_to();
}

//**************************************************************************
void corAssignFunction(corParserStruct& corparserStruct)
{
    corStackMark stack_mark;

    if (corexEnv->in_query_loop) {
        if (is_recursive(corparserStruct)) {
            fprintf(corexEnv->error_file, "CORAL::error :");
            fprintf(corexEnv->error_file,
                "Recursion not allowed in imperative rule!\n");
        }
        else {
            corcomplete_rule(corparserStruct.cur_rule, corparserStruct);

            corRelation *rel = corfind_external_relation(
                corparserStruct.cur_rule->head->pred,
                corparserStruct.cur_rule->head->arity());

            // Clear the lhs relation.
            if (rel) rel->empty_relation(0);

            // now execute the rule
            int num_ans = corexecute_single_rule(corparserStruct);

            if (corexEnv->C_verbose_answer_default) {
                fprintf(corexEnv->print_file,
                    "(Number of Tuples to Assign = %d)\n", num_ans);
            }
        }
    }
    else {
        fprintf(corexEnv->error_file, "CORAL::error :");
        fprintf(corexEnv->error_file, "Imperative statement inside module!\n");
    }
    if (corexEnv->C_verbose_answer_default) 
        fflush(corexEnv->print_file);

    stack_mark.pop_to();   
}

//**************************************************************************
void corQueryFunction(corParserStruct& corparserStruct)
{
    corStackMark stack_mark;
    corSymbol *varSymbol;
    corSymbol *headSymbol;

    if (!corexEnv->in_query_loop) {
        fprintf(corexEnv->error_file, "CORAL::error :");
        fprintf(corexEnv->error_file, "Imperative statement inside module!\n");
        return;
    }
    
#    ifdef DO_PROFILE
    corGlobalProfileInfo.clear() ;
#    endif

    if (corexEnv->C_answer_style_default == COR_EMPTY_ANSWER_PRINT)
        headSymbol = corPrintEmptySymbol;
    else if (corexEnv->C_answer_style_default == COR_TUPLE_ANSWER_PRINT)
        headSymbol = corPrintTupleSymbol;
    else
        headSymbol = corPrintBindingSymbol;

    corClause *cur_rule = corparserStruct.cur_rule;

    // make printa() the head of the rule 
    cur_rule->head = new corLiteral(corparserStruct, headSymbol,
        corparserStruct.rule_var_names.length);

    // create the arglist of printa() to have all the
    // variables in the body of the query 
    for (int i = 0; i < cur_rule->head->arity(); i++) {
        varSymbol=(corSymbol *)corparserStruct.rule_var_names.data[i];
        cur_rule->head->args[i] = new corVarArg(varSymbol,
            corrule_var_enter(corparserStruct, varSymbol));
    }
//    delete &(cur_rule->head->args[0]); xxx
    corexEnv->C_interactive_mode_current = corexEnv->C_interactive_mode_default;

    corcomplete_rule(corparserStruct.cur_rule, corparserStruct);

    int num_ans = corexecute_single_rule(corparserStruct);
    if (corexEnv->C_verbose_answer_default) {
        fprintf(corexEnv->print_file, "(Number of Answers = %d)\n", num_ans);
        fflush(corexEnv->print_file);
    }

    stack_mark.pop_to();
}

//**************************************************************************
void corFactFunction(corParserStruct& corparserStruct)
{
    if (corparserStruct.rhspreds.length > 0) {
        fprintf(corexEnv->error_file, "CORAL::error : ");
        fprintf(corexEnv->error_file, "Rule declared outside a module!\n");
    }

    corRelation *rel;
    rel = corfind_external_relation(corparserStruct.cur_rule->head->pred,
        corparserStruct.cur_rule->head->arity());

    // use the kludge that ensures that all builtins have count = -1
        
    if (!(corexEnv->C_insert_mode_default) && !corparseStack->is_empty() &&
        (corparseStack->top_file() == corcurConn->conn_sock->in))
    {
        if (!rel || rel->count >= 0) {
            fprintf(corexEnv->error_file, "CORAL::error : ");
            fprintf(corexEnv->error_file, "Not in insert mode !\n");
            return;
        }
    }
    
    if (!rel) {
        rel = cormake_relation(corparserStruct.cur_rule->head->pred,
            corparserStruct.cur_rule->head->arity());
    }
    
    corBindEnv *insert_env = NULL;
    if (corparserStruct.rule_var_names.length > 0)
        insert_env = new corVersionedBindEnv(corparserStruct.rule_var_names.length);
    
    rel->insert_new(corparserStruct.cur_rule->head->args, insert_env);
    if (insert_env)
        delete insert_env;
}

//**************************************************************************
void corEndModuleFunction(corParserStruct& corparserStruct)
{
    if (corparserStruct.CurModule.PreProcessing)
        corperform_rewriting(corparserStruct);
    else {
        corinterpret_module(corparserStruct) ;
    }
    corparserStruct.cur_rule = NULL;
}

//**************************************************************************
static int is_recursive(corParserStruct &corparserStruct)
{
    corLiteral *head = corparserStruct.cur_rule->head;
    corLiteral *body_pred;

    for (int i = 0; i < corparserStruct.rhspreds.length; i++) {
        body_pred = (corLiteral*)corparserStruct.rhspreds.data[i] ;
        if ((head->pred == body_pred->pred) &&
            (head->arity() == body_pred->arity()))
        {
            return 1;
        }
    }
    return 0;
}

//**************************************************************************
