//**************************************************************************
//      ParserStruct.h
//**************************************************************************

#include "ParserStruct.h"

#include "Predicate.h"
#include "Literal.h"

//**************************************************************************
void corParserStruct::init()
{
    // Many of these may need to have destructors called on them
    label_counter = 1;

    AllPredicates = NULL;
    Predicates = NULL;
    rule_list = NULL;
    make_index_annotations = NULL;
    agg_sel_annotations = NULL;
    prioritize_annotations = NULL;
    AllowedAdornList = NULL;
    DisallowedAdornList = NULL;
    MultisetAnnotations = NULL;
    diskrel_annotations = NULL;
    PredicateTable = NULL;
    CurModule.init();
    func_setup = 0;
    cur_arg_chain = NULL;
    //if (parameters)
    //    delete parameters;
    parameters = new corparameter_chain;
    cur_set = NULL;
    lhs_flag = 0;
    nested_index = 0;
    cur_rule = NULL;
    snrule = NULL;
    table = NULL;
    rInfo_arr = NULL;
}

//**************************************************************************
corParserStruct::~corParserStruct()
{
    clear(0);
}

//**************************************************************************
void corParserStruct::clear(int flag)
{
    if (!flag) {
        if (cur_rule) {
            if (cur_rule->head) {
                if (cur_rule->head->predicate)
                    delete cur_rule->head->predicate;
                delete cur_rule->head;
            }
            delete cur_rule;
        }
        cur_rule = NULL;
        if (PredicateTable){
            delete PredicateTable;
        }
        if (snrule) {
            delete snrule;
            delete table;
            delete rInfo_arr;
        }
        if (parameters){
            delete parameters;
            parameters = NULL;
        }
      }

    else 

      {
        // How Much of this should be commented out ?? : PRAVEEN
        if (cur_rule) {
            /*****
            if (cur_rule->head) {
                if (cur_rule->head->predicate)
                    delete cur_rule->head->predicate;
                delete cur_rule->head;
            }
            *****/
            delete cur_rule;
        }
        cur_rule = NULL;
        /****
        if (PredicateTable)
            delete PredicateTable;
        if (snrule) {
            delete snrule;
            delete table;
            delete rInfo_arr;
        }
        ****/
        if (parameters){
            delete parameters;
            parameters = NULL;
        }
      }
}

//**************************************************************************
void corParserStruct::reset(int flag)
{
    clear(flag);
    this->init();
}

//**************************************************************************
//      Find the (globally unique) Predicate with the given 'name' and 'arity'.
//      Allocate a new one if none is found.  LookupPredicate() is similar,
//      except that it does NOT allocate a new predicate if it is not found.
//**************************************************************************

corPredicate *corFindPredicate(corParserStruct & corparserStruct, Name name, int arity)
{
    corPredicate *pred;
    Association *ptr;

    if (corparserStruct.PredicateTable == NULL) {
        corparserStruct.PredicateTable = corAllocTab(COR_PREDICATE_TABLE_INCR, NULL);
    }
    ptr = corSymbolLookup(corparserStruct.PredicateTable, name);

    if (HashNone(ptr->arg)) {
        //      No previous Predicate with that name
        pred = corAllocPredicate(corparserStruct, name, arity);
        ptr->arg = (Name) name;
        ptr->val = pred;
        corTabInserted(corparserStruct.PredicateTable, ptr);
        pred->symtabnext = corparserStruct.Predicates;
        corparserStruct.Predicates = pred;
        return pred;
    }

    //  Check all existing Predicates with same 'name' for given 'arity'
    pred = (corPredicate *) ptr->val;
    for (;; pred = pred->symtabnext) {
        if (pred->arity == arity)
            return pred;
        else if ((pred->symtabnext == NULL)
                                || (pred->symtabnext->name != name)) {
            /* No such Predicate. Allocate and link in a new one. */
            corPredicate *npred = corAllocPredicate(corparserStruct, name, arity);
            npred->symtabnext = pred->symtabnext;
            pred->symtabnext = npred;
            return npred;
        }
    }
}

//**************************************************************************
//      Same as FindPredicate, except that if it is not found, a new predicate
//      is NOT created.  Instead NULL is returned.
//**************************************************************************

corPredicate *corLookupPredicate(corParserStruct & corparserStruct, Name name, int arity)
{
    corPredicate *pred;
    Association *ptr;

    if (corparserStruct.PredicateTable == NULL) {
        corparserStruct.PredicateTable = corAllocTab(COR_PREDICATE_TABLE_INCR, NULL);
    }
    ptr = corSymbolLookup(corparserStruct.PredicateTable, name);
    if (HashNone(ptr->arg)) {
        return NULL;
    }

    //      Check all existing Predicates with same 'name' for given 'arity'
    pred = (corPredicate *) ptr->val;

    for (;; pred = pred->symtabnext) {
        if ((arity == -1) || (pred->arity == arity))
            return pred;
        else if (pred->symtabnext == NULL || pred->symtabnext->name != name) {
            //      No such Predicate
            return NULL;
        }
    }
}

//**************************************************************************
