/*****************************************************************************/
//    HeapNode.C
/*****************************************************************************/

#include "HeapNode.h"

#include "CoralError.h"
#include "Assert.h"
#include "SymbolArg.h"

/*****************************************************************************/
corHeapNode::corHeapNode(corTuple *tuple1, corCArg val1)
{
    tuple = tuple1;
    val = val1;
    left = NULL;
    right = NULL;
    weight = 1;
}

/*****************************************************************************/
static void swap_values(corHeapNode *left, corHeapNode *right)
{
    corTuple *tmp_tuple = left->tuple;
    corCArg tmp_val = left->val;
    left->tuple = right->tuple;
    left->val = right->val;
    right->tuple = tmp_tuple;
    right->val = tmp_val;
}

/*****************************************************************************/
void cormerge_heaps(HeapNodePtr *rootptr, corHeapNode *left, corHeapNode *right,
    int order)
{
    if (!left) {
        *rootptr = right;
        return;
    }
    else if (!right) {
        *rootptr = left;
        return;
    }

    if (left->weight > right->weight) {
        if (left->val.compare_num(right->val) * order < 0)
            swap_values(left, right);
        left->weight += right->weight;
        cormerge_heaps( &(left->right), left->right, right, order);
        (*rootptr) = left;
    }
    else {
        right->weight += left->weight;
        if (right->val.compare_num(left->val) * order < 0)
            swap_values(left, right);
        cormerge_heaps( &(right->left), left, right->left, order);
        (*rootptr) = right;
    }
}

/*****************************************************************************/
corTuple* corpop_heap(HeapNodePtr *heap, int order) 
{
    corTuple *tuple1;

    if (! (*heap)) {
        CORAL_error( COR_INTERNAL_ERROR, "Attempt to pop empty heap!",
            "pop_heap(...)");
        return NULL;
    }

    tuple1 = (*heap)->tuple;
    ASSERT((tuple1!=NULL));

    corHeapNode *tmp = (*heap);
    cormerge_heaps(heap, (*heap)->left, (*heap)->right, order);
    delete tmp;

    return tuple1;
}

/*****************************************************************************/
void corheap_insert(corHeapNode *heap, corTuple *tuple1, corCArg val1, int order) 
{
    heap->weight ++;

    if ( (val1.compare_num(heap->val) * order) > 0 ) {
        corCArg tmp = heap->val;
        heap->val = val1;
        val1 = tmp;
        corTuple *tmptup = heap->tuple;
        heap->tuple = tuple1;
        tuple1 = tmptup;
    }

    if (!heap->left) {
        heap->left = new corHeapNode(tuple1, val1);
    }
    else if (!heap->right) {
        heap->right = new corHeapNode(tuple1, val1);
    }
    else if (heap->left->weight > heap->right->weight) {
        corheap_insert(heap->right, tuple1, val1, order);
    }
    else {
        corheap_insert(heap->left, tuple1, val1, order);
    }
}

/*****************************************************************************/

