/*****************************************************************************/
//    RMark.h
/*****************************************************************************/

#ifndef _RMark_H_
#define _RMark_H_

#include "TupleList.h"
//#include "RelationList.h"

/*****************************************************************************/

class corRelationLink;

/*****************************************************************************/
/*
    RMarks are used to select a sub-range of a relation.
    The range (RM1, RM2) of relation R is the set of tuples in R that
    were added to R after RM1 was created but before RM2 was created.
    Either RM1 or RM2 can be NULL, which is equivalent to beginning/end 
    of relation.  An RMark is created using 'relation->getMark()'.
*/
/*****************************************************************************/

class corRMark {
public:
    corRMark *prev;
    corRMark *next;
    corRelationLink *rel_p;        //    new for purposes of delta_indexing
    //    Given a startmark and an endmark, we consider each relation
    //    in Relation.delta_relations (which is a RelationList)
    //    starting from (and including) that associated with 
    //    startmark.rel_p, up to (but not including) that associated
    //    with endmark.rel_p
    
    corTupleLink **lastptr;
    //    (*lastptr) points to a TupleLink.  The semantics of 
    //    start and end marks is that tuples after (but not including)
    //    the tuple of *(startmark.lastptr) are to be included, up to 
    //    (and including) the tuple of *(endmark.lastptr).
    //    If *(startmark.lastptr) == *(endmark.lastptr), no tuples
    //    are included.

    corRMark();
    corRMark(corTupleLink **);
    corRMark(corRelationLink *);
    void init(corTupleLink **);
    void init(corRelationLink *);
};

/*****************************************************************************/

#endif _RMark_H_

