/*****************************************************************************/
//    Annotations.C
/*****************************************************************************/

#include "Annotations.h"

#include "VarArg.h"
#include "PredAnnotations.h"
#include "StackMark.h"
#include "ArrayBindEnv.h"
#include "Globals.h"
#include "ParserStruct.h"
#include "Compile.h"
#include "GenericIndex.h"
#include "Unparse.h"
#include "Unify.h"

#include "Interface.h"
#include "Util.h"
#include "CArg.h"

/*****************************************************************************/
int corduplicate_index(corPredAnnotations *annos, corSymbol *pred,
    corBitVector *bv,corArgList *al1, int num_vars, corArgList *al2)
{
    corStackMark stackmark;
    char buf[500];

    for (; annos; annos = annos->next()) {
        sprintf(buf, "%s", corSymbolString(annos->pred));
        if (annos->adorn)
            sprint_adornment(buf+strlen(buf), annos->adorn);

        if (((pred == annos->pred) || (pred == corEnterSymbol(buf))) &&
            ((bv == annos->adorn) ||
            (bv && annos->adorn && (*bv == *(annos->adorn)))))
        { 
            if (!al2 && !annos->arglist2) {
                if ((*al1)[0] == ((*(annos->arglist1))[0])) return 1;
            }
            else if (al2 && annos->arglist2) {
                corArrayBindEnv pat_env1(num_vars);
                corArrayBindEnv pat_env2(num_vars);
                corArrayBindEnv pat_env3(num_vars);
                corArrayBindEnv pat_env4(num_vars);

                //    check if pattern and the var_names of each subsume 
                //    that of the other

                if ((corsubsumes_arg_lists(*concat_arglists(al1, al2),
                    *concat_arglists(annos->arglist1, annos->arglist2),
                    &pat_env1, &pat_env2) == COR_U_SUCCEED) &&
                    (corsubsumes_arg_lists(
                        *concat_arglists(annos->arglist1, annos->arglist2),
                         *concat_arglists(al1, al2),
                        &pat_env3, &pat_env4) == COR_U_SUCCEED))
                {
                    stackmark.pop_to();
                    return 1;
                }
                stackmark.pop_to();
            }
        }
    }

    // for good measure
    stackmark.pop_to();
    return 0;
}

/*****************************************************************************/
/*
    WARNING: The bitvector passed to this function must be on static storage.
    It is not copied out.
*/
void coradd_make_index_annotation(corPredAnnotations *& annos, corCArg *cp,
    corCArg *ca ,corArgList *a1, corArgList *a2) {
  Name p=NULL, a=NULL;
//  printf("ca cp\n");
  if (cp) 
    p = cp->SymbolP();
  if (ca) 
    a = ca->SymbolP();
  coradd_make_index_annotation(annos, p, a, a1, a2); // another function!
 };

void coradd_make_index_annotation(corPredAnnotations *& annos, Name pred,
			       Name adorn_string, corArgList *al1, corArgList *al2)
{
// printf("name name\n");
    corBitVector *bitv = NULL;

    if (adorn_string) {
        bitv = new corBitVector(SymbolLength(adorn_string));
        int i;
        char *ap;
        for(i=0,ap = corSymbolString(adorn_string);i<SymbolLength(adorn_string);
            i++, ap++)
        {
            if ( *ap == 'b') bitv->set(i);
        }
    }

    corVarLink *rest;
    char *argindex = NULL;
    int num_vars = 0;

    if (al2) {
        rest = al1->var_list(NULL);
        rest = al2->var_list(rest);
        num_vars = corcount_var_list(rest);
        argindex = corsimplify_index(al1, num_vars , al2);
    }

    if (argindex) {
        al1 = new corArgList(1);
        (*al1)[0] = corEnterSymbol(argindex);
        al2 = NULL;
        delete [] argindex;
    }

    if (corduplicate_index(annos, pred, bitv, al1, num_vars, al2)) {
        return;
    }
    
    corPredAnnotations *pa = new corPredAnnotations (pred, annos);
                 
    pa->arglist1=al1;
    pa->arglist2=al2;
    pa->adorn = bitv;

    annos = pa;
}

/*****************************************************************************/
void coradd_make_index_annotation(corPredAnnotations*& annos, corSymbol *pred,
    corBitVector *bv, corArgList *al1, corArgList *al2)
{
    corVarLink *rest;
    char *argindex = NULL;
    int num_vars = 0;
//    printf("name bitvector\n");
    if (al2) {
        rest = al1->var_list(NULL);
        rest = al2->var_list(rest);
        num_vars = corcount_var_list(rest);
        argindex = corsimplify_index(al1, num_vars, al2);
    }

    if (argindex) {
        al1 = new corArgList(1);
        (*al1)[0] = corEnterSymbol(argindex);
        al2 = NULL;
        delete [] argindex;
    }

    if (corduplicate_index(annos, pred, bv, al1, num_vars, al2)) {
        return;
    }

    corPredAnnotations *pa = 
        new corPredAnnotations (pred, annos);
    pa->arglist1=al1;
    pa->arglist2=al2;
    pa->adorn = bv;

    annos = pa;
}


/*****************************************************************************/
void coradd_prioritize_annotation(corParserStruct& corparserStruct,corCArg *cPredName,
			       corCArg *cAdornStr, corArgList *al1, corCArg *op_arg)
{
    corSymbol *pred=NULL;
    corSymbol *adorn_string=NULL ;
    if (cPredName) 
      pred = cPredName->SymbolP();
    if (cAdornStr)
      adorn_string = cAdornStr->SymbolP();
    corPredAnnotations *pa =
        new corPredAnnotations(pred,corparserStruct.prioritize_annotations);

    pa->arglist1=al1;
    pa->arglist2=NULL;
    if (adorn_string) { // adornment specified
        int i;
        char *ap;
        pa->adorn = new corBitVector(SymbolLength(adorn_string));
        for(i=0,ap = corSymbolString(adorn_string); i<SymbolLength(adorn_string);
            i++, ap++)
        {
            if ( *ap == 'b') pa->adorn->set(i);
        }
    }
    else
        pa->adorn = NULL;

    pa->op_arg = *op_arg;
    corparserStruct.prioritize_annotations = pa;
    corparserStruct.CurModule.HasPrioritize = 1;
}

/*****************************************************************************/
void coradd_agg_sel_annotation(corParserStruct& corparserStruct, corCArg *cPredName,
			    corCArg *cAdornStr, corArgList *al1, corArgList *al2, 
			    corCArg *cOp_arg)
{
    corSymbol *pred=NULL;
    corSymbol *adorn_string=NULL ;
    if (cPredName) 
      pred = cPredName->SymbolP();
    if (cAdornStr)
      adorn_string = cAdornStr->SymbolP();
    
    corPredAnnotations *pa =
        new corPredAnnotations(pred, corparserStruct.agg_sel_annotations);

    pa->arglist1=al1;
    pa->arglist2=al2;
    if (adorn_string) {            // adornment specified
        int i;
        char *ap;
        pa->adorn = new corBitVector(SymbolLength(adorn_string));
        for (i=0,ap = corSymbolString(adorn_string); i<SymbolLength(adorn_string);
            i++, ap++)
        {
            if ( *ap == 'b') pa->adorn->set(i);
        }
    }
    else
        pa->adorn = NULL;

    pa->op_arg = *cOp_arg;
    corparserStruct.agg_sel_annotations = pa;
}

/*****************************************************************************/
void coradd_allowed_adornment(corParserStruct& corparserStruct, corCArg *c,
    corArgList* args)
{
 Name pred_name=NULL;
 if (c) pred_name = c->SymbolP();
    for (int i=0;i < args->count();i++) {
        if ((*args)[i].kindof() != COR_SYMBOL_CONST) {
            fprintf(corexEnv->error_file,
                "Bad kind of adornment for allowed adorns %s.\n",
                pred_name->string());
            continue;
        }
        Name adornname = (*args)[i].SymbolP();

        coradd_allowed_adornment(corparserStruct, pred_name, 
			      adornname->string(),
			      adornname->length());
    }
}

/*****************************************************************************/
void coradd_allowed_adornment(corParserStruct& corparserStruct, Name pred_name,
    char* adorn_string, int adorn_length)
{
    char *temp_adorn_string = adorn_string;

    for (int i=0; i < adorn_length;i++) {
        char adorn_char = *temp_adorn_string++;
        switch (adorn_char) {
        case 'b':
            break;
        case 'f':
            break;
        case 'a':
        {
            char *new_adorns = new char[adorn_length+1];
            strcpy(new_adorns, adorn_string);
            new_adorns[i] = 'b';
            coradd_allowed_adornment(corparserStruct, pred_name, new_adorns,
                adorn_length);
            new_adorns[i] = 'f';
            coradd_allowed_adornment(corparserStruct, pred_name, new_adorns,
                adorn_length);
            delete [] new_adorns;
            return;
        }
        default:
            fprintf(corexEnv->error_file,
                "Bad character in adornment string '%s'\n",
                adorn_string);    
            return;
        }
    }

    // should get here only if all the characters in adorn_string are
    // either 'b' or 'f'

    corBitVector *bv = new corBitVector();

    corConvertAdornment(*bv, adorn_string, adorn_length); 
    corPredAnnotations *pa = new corPredAnnotations (pred_name,
        corparserStruct.AllowedAdornList);
    pa->arglist1 = pa->arglist2 = NULL;
    pa->adorn = bv;
    corparserStruct.AllowedAdornList = pa;
}

/*****************************************************************************/
void coradd_disallowed_adornment(corParserStruct& corparserStruct, Name pred_name,
    corArgList* args)
{
    for (int i=0;i < args->count();i++) {
        if ((*args)[i].kindof() != COR_SYMBOL_CONST) {
            fprintf(corexEnv->error_file,
                "Bad kind of adornment for disallowed adorns %s.\n",
                pred_name->string());
            continue;
        }
        Name adornname = (Name)((*args)[i].other);

        coradd_disallowed_adornment(corparserStruct, pred_name, adornname->string(),
            adornname->length());
    }
}

/*****************************************************************************/
void coradd_disallowed_adornment(corParserStruct& corparserStruct, Name pred_name,
    char* adorn_string, int adorn_length)
{
    char *temp_adorn_string = adorn_string;

    for (int i=0; i < adorn_length;i++) {
        char adorn_char = *temp_adorn_string++;
        switch (adorn_char) {
        case 'b':
            break;
        case 'f':
            break;
        case 'a': 
        {
            char *new_adorns = new char[adorn_length+1];
            strcpy(new_adorns, adorn_string);
            new_adorns[i] = 'b';
            coradd_disallowed_adornment(corparserStruct, pred_name, new_adorns,
                adorn_length);
            new_adorns[i] = 'f';
            coradd_disallowed_adornment(corparserStruct, pred_name, new_adorns,
                adorn_length);
            delete [] new_adorns;
            return;
        }
        default:
            fprintf(corexEnv->error_file,
                "Bad character in adornment string '%s'\n",
                adorn_string);    
            return;
        }
    }

    // should get here only if all the characters in adorn_string are
    // either 'b' or 'f'

    corBitVector *bv = new corBitVector();

    corConvertAdornment(*bv, adorn_string, adorn_length); 
    corPredAnnotations *pa = new corPredAnnotations (pred_name,
        corparserStruct.DisallowedAdornList);
    pa->arglist1 = pa->arglist2 = NULL;
    pa->adorn = bv;
    corparserStruct.DisallowedAdornList = pa;
}

/*****************************************************************************/
void coradd_multiset_annotation(corParserStruct& corparserStruct, corCArg* c,
    corArgList *al1, int check_subsum)
{
 Name pred=NULL;
 if (c)
   pred = c->SymbolP();
    if (al1->count()) {
        FOR_EACH_ARG(arg,*al1) {
            if(arg->kindof() != COR_SYMBOL_CONST) {
                fprintf(corexEnv->error_file, 
                    "Error: Bad argument to multiset annotation for %s\n",
                    pred->string() );
            }
            else {
                corPredAnnotations *pa =
                    new corPredAnnotations(pred,corparserStruct.MultisetAnnotations);

                pa->arglist1=NULL;
                pa->check_subsum = check_subsum;
                corparserStruct.MultisetAnnotations = pa;
                corSymbol *adorn_string = arg->SymbolP();
                int i;
                char *ap;
                pa->adorn = new corBitVector(SymbolLength(adorn_string));
                for(i=0,ap = corSymbolString(adorn_string);
                    i<SymbolLength(adorn_string); i++, ap++)
                {
                    if ( *ap == 'b') pa->adorn->set(i);
                }
            }
        } END_EACH_ARG
    }
    else {
        corPredAnnotations *pa = new corPredAnnotations (pred,
               corparserStruct.MultisetAnnotations);
        pa->arglist1=al1;
        pa->adorn = NULL;
        pa->check_subsum = check_subsum;
        corparserStruct.MultisetAnnotations = pa;
    }
}

/*****************************************************************************/
void coradd_diskrel_annotation(corParserStruct& corparserStruct, corCArg *cPredName,
			    corCArg *cAdornStr, corArgList *al1)
{
    corSymbol *pred=NULL;
    corSymbol *adorn_string=NULL ;
    if (cPredName) 
      pred = cPredName->SymbolP();
    if (cAdornStr)
      adorn_string = cAdornStr->SymbolP();
 
    if (adorn_string) {
        corPredAnnotations *pa =
            new corPredAnnotations (pred, corparserStruct.diskrel_annotations);

        pa->arglist1=al1;
        corparserStruct.diskrel_annotations = pa;
        int i;
        char *ad_str = corSymbolString(adorn_string);
        pa->adorn = new corBitVector(strlen(ad_str));
        for (i=0; i < strlen(ad_str); i++) 
            if (ad_str[i] == 'b') pa->adorn->set(i);
    }
    else {
        corPredAnnotations *pa =
            new corPredAnnotations (pred, corparserStruct.diskrel_annotations);

        pa->arglist1=al1;
        pa->adorn = NULL;
        corparserStruct.diskrel_annotations = pa;
    }
}

/*****************************************************************************/
void corupdate_anno_list(Name name, corBitVector &adorn, corPredAnnotations *anno) 
{
    for (; anno != NULL; anno = anno->next()) {
        if (anno->pred == name && anno->adorn == NULL) {
            corPredAnnotations *newanno = new corPredAnnotations(anno->pred, NULL);
            *newanno = *anno;
            corBitVector *newvec = new corBitVector();
            *newvec = adorn;
            newanno->adorn = newvec;
            newanno->link(anno->next());
            anno->link(newanno);
        }
    }
}

/*****************************************************************************/
/*
    WARNING:  If more annotations are added, update this procedure
    as well as the printing procedures in unparse.C
*/

void corupdate_annotations(Name name, corBitVector& adorn)
{
    corupdate_anno_list(name, adorn, corparserStruct.MultisetAnnotations);
    corupdate_anno_list(name, adorn, corparserStruct.agg_sel_annotations);
    corupdate_anno_list(name, adorn, corparserStruct.make_index_annotations);
    corupdate_anno_list(name, adorn, corparserStruct.prioritize_annotations);
    corupdate_anno_list(name, adorn, corparserStruct.diskrel_annotations);
}

/*****************************************************************************/
void corannotation_error(corPredAnnotations *anno, char *msg)
{
    fprintf(corexEnv->error_file, "** Error in annotation: \n");
    anno->print(corexEnv->error_file);
    fprintf(corexEnv->error_file, "\n       -- %s.\n", msg);
}

/*****************************************************************************/
void coradd_index_delta_annotation(corParserStruct&, corCArg*, corArgList*)
{
    fprintf(corexEnv->error_file,
        "Sorry - add_index_delta_annotation not yet implemented\n");
}

/*****************************************************************************/
void coradd_index_delta_annotation()
{
    fprintf(corexEnv->error_file,
        "Sorry - add_index_delta_annotation not yet implemented\n");
}

/*****************************************************************************/
corRuleAnnoCaseDetails* cormake_anno_case_details( corCArg*)
{
    return NULL;
}

/*****************************************************************************/
corRuleAnnoCaseDetails* cormake_anno_case_details( ArgPtr )
{
    return NULL;
}

/*****************************************************************************/
corRuleAnnoCaseDetails* cormake_anno_case_details( corRuleAnnoCase *)
{
    return NULL;
}

/*****************************************************************************/
corRuleAnnoCaseName* cormake_anno_case_name(corCArg*)
{
    return NULL;
}

/*****************************************************************************/
corRuleAnnoCaseName* cormake_anno_case_name(ArgPtr)
{
    return NULL;
}

/*****************************************************************************/
corRuleAnnoCase* corlink_anno_case_list( corRuleAnnoCase*, corRuleAnnoCase*)
{
    return NULL;
}

/*****************************************************************************/
corRuleAnno* corlink_rule_anno(corRuleAnno*, corRuleAnno*)
{
    return NULL;
}

/*****************************************************************************/
corRuleAnnoCase* cormake_anno_case(corRuleAnnoCaseName*, corRuleAnnoCaseDetails*)
{
    return NULL;
}

/*****************************************************************************/
corRuleAnno* cormake_anno(corCArg*, corRuleAnnoCaseDetails*)
{
    return NULL;
}

/*****************************************************************************/

