/*****************************************************************************/
//    QFModuleData.h
/*****************************************************************************/

#ifndef _QFModuleData_H_
#define _QFModuleData_H_

#include "RuleInfoArray.h"
#include "ModulePred.h"
#include "ModuleInfo.h"
#include "PredAnnotations.h"
#include "Predicate.h"
#include "RuleInfo.h"
#include "RelationMisc.h"
#include "SemiNaive.h"        // ModuleKind

/*****************************************************************************/

class corTable;
class corTupleIterator;

/*****************************************************************************/
// For each exported query form in a user-defined CORAL module, there is
// one QFModuleData structure created. The Solve() method provides a
// mechanism to evaluate the rules defined in the module. Each such
// structure is associated with its exported query form in a table
// of DerivedMethods (see derived-rel.h). QFModuleData is a generic
// class with subclasses for Materialzed and Pipelined modules.
// Sudarshan: This is a static data structure and should not contain
// any fields which differ across different calls to a module
/*****************************************************************************/

class corQFModuleData {
public:
    corTable* F;                        // Table for the query
    corRuleInfoArray* ruleArray;        // Array of rules for the query form

    corModulePred* predicates;            // Used for run_time initialisation
                                    // of rule pointers
    corModuleInfo module_info;
    corPredAnnotations* index_annotations;
    corPredAnnotations* multiset_annotations;
    corPredAnnotations* agg_sel_annotations;
    corPredAnnotations* prioritize_annotations;
    corPredAnnotations* diskrel_annotations;

    corPredicate* export_pred;            // Predicate on which query is asked

    RelationOptions rel_options;
    long display_rules;

    corRuleInfo* curr_rInfo;

    corTable* free_table_list;            // List of (initialized) table structs
                                    // that can be reused.

    virtual void Solve(corTable *F, corTupleIterator *iterator = NULL) = 0;
    virtual ModuleKind kindof() = 0;
};

/*****************************************************************************/

#endif _QFModuleData_H_

