/*****************************************************************************/
//    Tuple.h
/*****************************************************************************/

#ifndef _Tuple_H_
#define _Tuple_H_

#include "BindEnv.h"
#include "ArgList.h"
#include "Term.h"
#include "ObjectCount.h"

/*****************************************************************************/
// A Tuple has an ArgList and a BindEnv to interpret the ArgList in.
// If the BindEnv is NULL, then we assume that all variables in the tuple are
// free.  

class corTuple {
public:
    unsigned int deleted : 1;
    corBindEnv *bindenv;  
    corArgList* _args;        // array of terms. Length==relation's arity
    int env_size;        // Gives number of variables in the tuple---
                        // WARNING:  set this to -1 if it's not known

    int contid;            //  This field and the next are related to
    int parid;            //  return unification optimizations.

    int ref_count;      // tuple reference count

    
    // Used to be copy_env
    corBindEnv *make_bindenv_version();       
                // Returns a copy of the bindenv, but default
                // bindenvs result in full copies, not default
                // copies.
        
    corTuple() {
        _args = NULL;
        bindenv = NULL;
        env_size = 0;
        deleted = 0;
        contid = parid = ref_count = 0;
#        ifdef DEBUG
            corC_NewTupleCount ++;
            corC_NewTupleSize += sizeof(corTuple);
#        endif
    }
    corTuple(int arity) {
        _args = new corArgList(arity);
        bindenv = NULL;
        env_size = 0;
        deleted = 0;
        contid = parid = ref_count = 0;
#        ifdef DEBUG
            corC_NewTupleCount ++;
            corC_NewTupleSize += sizeof(corTuple); 
#        endif
    }

    corTuple(corArgList *arglist, corBindEnv *env) {
        _args = arglist;
        bindenv = env;
        env_size = 0;
        deleted = 0;
        contid = parid = ref_count = 0;
#        ifdef DEBUG
            corC_NewTupleCount ++;
            corC_NewTupleSize += sizeof(corTuple);
#        endif
    }

    corTuple(corArgList&a) : _args(&a) {
        bindenv = NULL;
        env_size = 0;
        deleted = 0;
        contid = parid = ref_count = 0;
#        ifdef DEBUG
            corC_NewTupleCount ++;
            corC_NewTupleSize += sizeof(corTuple);
#        endif
       }

    corTuple(corArgList *arglist) {
        _args = arglist;
        bindenv = NULL; 
        env_size = 0;
        deleted = 0;
        contid = parid = ref_count = 0;
#        ifdef DEBUG
            corC_NewTupleCount ++;
            corC_NewTupleSize += sizeof(corTuple);
#        endif
    }

    ~corTuple() {
#        ifdef DEBUG
            corC_DeleteTupleCount++;
            corC_DeleteTupleSize += sizeof(corTuple);
#        endif
    }

    // true if all vars are free, so no bindenv is allocated.
    inline int default_bindenv() const { return (bindenv == NULL); }

    // True if all arguments are ground.
    int isConstant();
    inline void do_delete()        { deleted = 1; }
    inline int is_deleted()        { return deleted; }
    inline corArgList& args()        { return *_args; }
    inline int arity() const    { return _args->count(); }

    // Returns a term that is the i'th argument of
    // the tuple.
    inline corTerm *term(int i) const {
        corTerm *term = new corTerm ((*_args)[i], bindenv);
        return term;
    }

    inline corCArg& arg(int i) const { return (*_args)[i]; }
    corCArg& operator [] (int i) const;  // Cannot make this inline !! - Praveen
    corTuple * rename(corBindEnv *env);  // Renames variables in tuple to start from
        // after the highest var in env;  also adds new variables to
        // env

    inline int hash( corBitVector& bv ) const {
        return _args->hash( bv, bindenv );
    }

    inline int hash() const {
        return _args->hash(bindenv);
    }
    void printon(FILE *file) const;
    void update1(int argnum, corArgList *arg);
};

/*****************************************************************************/

#endif _Tuple_H_

