// *************************************************************************
//    Print.C
// *************************************************************************

#include "Print.h"

#include "Globals.h"
#include "Rule.h"
#include "Literal.h"
#include "DontCareArg.h"
#include "Unparse.h"
#include "ExportInfo.h"

/*****************************************************************************/

static void print_status(corClause *rule, FILE *out);

/*****************************************************************************/
void print_bits(corBitVector corb, int flag, FILE *outf) 
{
    register int i;
    char c[32];
    char ch;

    if (corb.length == 0) return;
    fprintf(outf, "_");
    ch = (flag==0) ? 'd':'_';

    for (i=0; i<corb.length; ++i)
        c[i] = corb[i] ? ch : 'n';

    c[corb.length]='\0';

    fprintf(outf, "%s", c);
}

/*****************************************************************************/
void print_bits(corBitVector corb, int flag)
{
    print_bits(corb, flag, stderr);
}

/*****************************************************************************/
static void print_status(corClause *rule, FILE *out)
{
    char *str;
    switch(rule->factor_status) {
        case 0 : str = "COR_NON_RECURSIVE   "; break;
        case 1 : str = "COR_RT_LINEAR       "; break;
        case 2 : str = "COR_LT_LINEAR       "; break;
        case 3 : str = "PSEUDO_LT_LINEAR    "; break;
        case 4 : str = "COR_MULTI_LINEAR    "; break;
        default : str = "COR_NON_LINEAR     "; break;
    }
    fprintf(out, "%s : ", str);
}

/*****************************************************************************/
void print_factor(struct corcollection *factor_rules, FILE *outf,
     int show_status,struct corcollection *org_rules )
{
    if ((corClause *)(factor_rules->chain) == NULL) {
        fprintf(stderr,
   "CORAL:: WARNING! Magic/Factoring for %s results in module with no rules\n",
                 corparserStruct.CurModule.name->string());
        return;
    }

    if (corparserStruct.CurModule.name) {
        fprintf(outf, "module %s.\n",
            corSymbolString(corparserStruct.CurModule.name));
    }
    for (corExportInfo *export_info = corparserStruct.CurModule.ExportedPreds;
        export_info != NULL; export_info = export_info->next)
    {
        fprintf(outf, "export %s[", export_info->pred_name->string());
        if (export_info->adorn_symbol) {
            fputs(export_info->adorn_symbol->string(), outf);
        }
        fprintf(outf, "].\n");
    }

    corprint_Annotations(corparserStruct.CurModule, outf,
        corparserStruct.make_index_annotations, corparserStruct.agg_sel_annotations,
        corparserStruct.AllowedAdornList, corparserStruct.MultisetAnnotations,
        corparserStruct.prioritize_annotations, corparserStruct.diskrel_annotations);

    FOR_EACH_RULE(factorrule, factor_rules) {
        if (show_status)
            print_status(factorrule, outf);
        corprint_rule(factorrule, outf);
    } END_EACH_RULE

    if (org_rules) { // added by Ken to print orginal rules
      fprintf(outf, "%% original program:\n");

      FOR_EACH_RULE(orgrule, org_rules) {
          if (show_status)
              print_status(orgrule, outf);
          orgrule->sips = NULL;
          corprint_rule(orgrule, outf);
      } END_EACH_RULE
    }

    fprintf(outf, "end_module.\n\n");
}

/*****************************************************************************/

