
%{
/*****************************************************************************/
/*    Lexer.l   */
/*****************************************************************************/

#include <string.h>

#define TTLEX_EXTRN
#include "Lexer.h"

#undef ttwrap
int ttwrap();

#ifndef MAX_STR_LEN
#define MAX_STR_LEN    1024
#endif

static char  str_buff[MAX_STR_LEN];
static char *str_end = &str_buff[MAX_STR_LEN-1];
static char *str_ptr;

int yylineno = 1;
int in_comment = 0;
int in_embed = 0;
int in_proto = 0;

extern FILE *error_file;
extern char *input_file_name;

#ifdef FLEX_SCANNER

#undef YY_INPUT

    /*  Scan input from a file pointer (default)  */
#define YY_INPUT(buf,result,max_size) \
    { \
        if (fgets(buf,max_size,yyin) == NULL) { \
           result = YY_NULL; \
        } else { \
           result = strlen(buf); \
        } \
    }

#else    /* not FLEX_SCANNER  */

static int input_char;

#undef input
#undef unput

    /*  Scan input from a file pointer (default)  */
#define input() \
     ((((input_char=getc(yyin))==EOF) || (input_char == 0)) ? (0) : input_char)

#define unput(c) { ungetc(c, yyin); }

#endif     /* not FLEX_SCANNER  */


extern int yylex (YYSTYPE *lvalp);

#undef YYLMAX
#define YYLMAX 2048

/*  To give some args to yylex */
#undef YY_DECL
#define YY_DECL int yylex YY_PROTO((YYSTYPE *lvalp))

    /* This is strictly for prettiness, place something at same
       indentation as something else. */
#define INDENT(PROTO)    do { \
            char *_ptr = (PROTO); \
            while( *_ptr && ((*_ptr == ' ') || (*_ptr == '\t') \
                    || (*_ptr == '\f'))) \
               fputc(*_ptr++,yyout); \
        } while(0)

#define PUT  do { fputs(yytext,yyout); } while(0)

%}

%Start EMBED C_COMMENT C_STR EX_PROTO CPP_LINE
%a 9000
%o 9000

dot             \.
digit           [0-9]
alpha           [a-zA-Z_]
capalpha        [A-Z]
alphanum        {alpha}|{digit}
optsign         [\+\-]?
litquote        \\\"

number          {digit}+
integer         {optsign}{number}
frac            {dot}{number}
mantissa        {number}|{number}{frac}
exponent        [Ee]{integer}
real            {mantissa}{exponent}?

id              {alpha}{alphanum}*
string          \"([^\"]|{litquote})*\"
id_or_str       {id}|{string}

whitespace      [ \t\f]
ws              {whitespace}*
pp_strt         ^{ws}"#"{ws}

%%

<INITIAL>"\\["  { in_embed = 1; 
                  BEGIN(EMBED);
                  return(my_CORAL_BEGIN);
                }

<EMBED>"\\]"    { in_embed = 0;
                  BEGIN(INITIAL);
                  return(my_CORAL_END);
                }

<INITIAL,EMBED,EX_PROTO>"/*"  { PUT; BEGIN(C_COMMENT); in_comment = 1;}
<INITIAL,EX_PROTO>{pp_strt}[^\\\n]*\n  { PUT; yylineno++; }
<INITIAL,EX_PROTO>{pp_strt}[^\\\n]*    { PUT; BEGIN(CPP_LINE); }


<INITIAL,EX_PROTO>"//".*\n    { /* C++ style comment */
                                PUT;
                                yylineno++;
                              }
                    

<C_COMMENT>[^*\n]*      { PUT; /* ignore anything that's not a '*' */ }
<C_COMMENT>"*"+[^*/\n]* { PUT; /* ignore '*'s not followed by '/'s */ }
<C_COMMENT>\n           { PUT;
                          yylineno++;
                        }

<C_COMMENT>"*"+"/"      { 
                          PUT;
                          in_comment = 0;
                          if (in_embed)
                            BEGIN(EMBED);
                          else if (in_proto)
                            BEGIN(EX_PROTO);
                          else
                            BEGIN(INITIAL);
                        }


<INITIAL>{ws}"init_coral"{ws}"("[^\n]*     {

                      if (! strchr(yytext,';')){
                        fprintf(stderr,"Error at line %d, init_coral statement must be entirely on a single line.\n",yylineno);
                      }
                      if (yytext[yyleng-1] == '\n')
                        yylineno++;
                      fputs("\n",yyout);
                      INDENT(yytext);
                      fputs("corexEnv->C_exec_mode = COR_DOING_IMPERATIVE;\n",
                        yyout);
                      PUT;
                      fprintf(yyout,"\n");
                      INDENT(yytext);
                      return(my_INIT_CORAL);
                    }

<INITIAL>{ws}"_coral_export"{ws} {
                      BEGIN(EX_PROTO);
                      in_proto = 1;
                      return(my_EXPORT_CORAL);
                    }

<EX_PROTO>";"       {
                      PUT;
                      in_proto = 0;
                      BEGIN(INITIAL);
                      return(SEMICOLON);
                    }

<EX_PROTO>"char"    {
                      PUT;
                      return(CHAR);
                    }

<EX_PROTO>"int"     {
                      PUT;
                      return(INT);
                    }

<EX_PROTO>"float"   {
                      PUT;
                      return(FLOAT);
                    }

<EX_PROTO>"double"  {
                      PUT;
                      return(DOUBLE);
                    }

<EX_PROTO>"void"    {
                      PUT;
                      return(VOID);
                    }

<EX_PROTO>"char"{ws}*"*"  {
                      PUT;
                      return(CHAR_PTR);
                    }

<EX_PROTO>"Arg"     {
                      PUT;
                      return(ARG);
                    }

<EX_PROTO>"class"   {
                      PUT;
                      return(CLASS);
                    }

<EX_PROTO>"struct"  {
                      PUT;
                      return(STRUCT);
                    }

<EX_PROTO>"short"   {
                      PUT;
                      return(SHORT);
                    }

<EX_PROTO>"long"    {
                      PUT;
                      return(LONG);
                    }

<EX_PROTO>"signed"  {
                      PUT;
                      return(SIGNED);
                    }

<EX_PROTO>"unsigned" {
                      PUT;
                      return(UNSIGNED);
                    }

<EX_PROTO>{id}      {
                      PUT;
                      lvalp->val.str = strdup(yytext);
                      return(IDENT);
                    }

<EX_PROTO>{whitespace}+ { /* space/tab/formfeed; ignore */
                          PUT;
                        }

<EX_PROTO>\n|\r     {
                      yylineno++;
                      PUT;
                    }

<C_STR>\"           { /* Closing quote */
                       *str_ptr++ = '"';
                       *str_ptr = '\0';
                        BEGIN(EMBED);
                        lvalp->val.str = strdup(str_buff);
                        return(STRING);
                    }

<C_STR>[^\\\"]+     { /* Normal chars */
                        if ((str_ptr + yyleng) < str_end){
                            strcpy(str_ptr,yytext);
                            str_ptr += yyleng;
                        }
                    }

<C_STR>\\.          { /* Literal char (like " and \) */
                        *str_ptr++ = '\\';
                        if (str_ptr < str_end)
                            *str_ptr++ = yytext[1];
                    }

<CPP_LINE>\\\n      { PUT; yylineno++; }
<CPP_LINE>\\.       { PUT; }
<CPP_LINE>.         { PUT; }
<CPP_LINE>\n        { PUT; yylineno++;
                      if (in_proto)
                        BEGIN(EX_PROTO);
                      else
                        BEGIN(INITIAL);
                    }

<INITIAL>{whitespace}+  { PUT; }

<INITIAL>.|\n       {
                      if (*yytext == '\n')
                        yylineno++;
                      PUT;
                    }

<EMBED>\"[^\\\"]*\" {   /* Simple quoted strings */
                        lvalp->val.str = strdup(yytext);
                        return(STRING);
                    }

<EMBED>\"         { /* For strings that contain escaped characters */
                      BEGIN(C_STR);
                      str_ptr = str_buff;
                     *str_ptr++ = '"';
                  }

<EMBED>"char"     { return(CHAR); }
<EMBED>"int"      { return(INT); }
<EMBED>"float"    { return(FLOAT); }
<EMBED>"double"   { return(DOUBLE); }
<EMBED>"void"     { return(VOID); }
<EMBED>"char"{ws}*"*"  { return(CHAR_PTR); }
<EMBED>"string"   { return(CHAR_PTR); }
<EMBED>"Arg"      { return(ARG); }
<EMBED>"class"    { return(CLASS); }
<EMBED>"struct"   { return(STRUCT); }
<EMBED>"short"    { return(SHORT); }
<EMBED>"long"     { return(LONG); } 
<EMBED>"signed"   { return(SIGNED); } 
<EMBED>"unsigned" { return(UNSIGNED); }

<EMBED>"sips"     { return(SIPS); }
<EMBED>"alias"    { return(ALIAS); }
<EMBED>"mod"      { return(MOD); }
<EMBED>"not"      { return(NOT); }
<EMBED>"del"      { return(DEL); }

<EMBED>"c_module" { 
                      return(START_C_MODULE);
                  }

<EMBED>"end_c_module" { 
                            return(END_C_MODULE);
                      }
<EMBED>"module"     { return(MODULE); }
<EMBED>"export"     { return(EXPORT); }
<EMBED>"consult"    { /* Special return value */ 
                        lvalp->val.str = strdup(yytext);
                        return(CONSULT);
                    }
<EMBED>"end_module" { return(END_MODULE); }

<EMBED>"<<"       { return(LSHIFT); }
<EMBED>">>"       { return(RSHIFT); }
<EMBED>"bitxor"   { return(BITXOR); }
<EMBED>"bitand"   { return(BITAND); }
<EMBED>"bitor"    { return(BITOR); }
<EMBED>"bitnot"   { return(BITNOT); }  
<EMBED>"is"       { return(EQUAL); }

<EMBED>[A-Z]{alphanum}*  |
<EMBED>[A-Z][A-Za-z0-9_.]*{alphanum}+    {
                            lvalp->val.str = strdup(yytext);
                            return(VAR);
                    }

<EMBED>"_"          {
                        lvalp->val.str = strdup(yytext);
                        return(VAR);
                    }

<EMBED>{number}     {
                        lvalp->val.inum = atoi(yytext);
                        return(INUM);
                    }

<EMBED>{real}       {
                        lvalp->val.rnum = atof(yytext);
                        return(RNUM);
                    }


<EMBED>[a-z]([#A-Za-z0-9_.]*[#A-Za-z0-9_])? {
                        lvalp->val.str = strdup(yytext);
                        return(SYMBOL);
                    }


<EMBED,EX_PROTO>","      { return(COMMA); } 
<EMBED,EX_PROTO>":"      { return(COLON); }
<EMBED,EX_PROTO>"."      { return(PERIOD); }
<EMBED>":-"     { return(MARK); }
<EMBED>"<-"     { return(MARK); }
<EMBED>"+="     { return(APPEND); }
<EMBED>"-="     { return(DELETE); }
<EMBED>"*="     { return(UPDATE); }
<EMBED>":="     { return(ASSIGN); }
<EMBED>"==>"    { return(ARROW); }
<EMBED>"//"     { return(SLASHSLASH); }
<EMBED>"("      { return(LPAREN); }
<EX_PROTO>"("   { PUT; return(LPAREN); }
<EMBED>")"      { return(RPAREN); }
<EX_PROTO>")"   { PUT; return(RPAREN); }
<EMBED>"{"      { return(LBRACE); }
<EX_PROTO>"{"   { PUT; return(LBRACE); }
<EMBED>"}"      { return(RBRACE); }
<EX_PROTO>"}"   { PUT; return(RBRACE); }
<EMBED>"["      { return(LBRACKET); }
<EX_PROTO>"["   { PUT; return(LBRACKET); }
<EMBED>"]"      { return(RBRACKET); }
<EX_PROTO>"]"   { PUT; return(RBRACKET); }
<EMBED,EX_PROTO>"$"      { return(DOLLAR); }
<EMBED>"|"      { return(VBAR); }
<EMBED>"@"      { return(AT); }
<EMBED,EX_PROTO>"+"      { return(PLUS); }
<EMBED,EX_PROTO>"-"      { return(MINUS); }
<EMBED,EX_PROTO>"*"      { return(STAR); }
<EMBED,EX_PROTO>"/"      { return(SLASH); }
<EMBED>"?"      { return(QUESTION); }
<EMBED,EX_PROTO>"="      { return(EQUAL); }
<EMBED>"\\="    { return(NOTEQUAL); }
<EMBED>"~="     { return(NOTEQUAL); }
<EMBED>"<"      { return(LESS); }
<EMBED>">"      { return(GREATER); }
<EMBED>"<>"     { return(NOTEQUAL); }
<EMBED>"<="     { return(LESSEQUAL); }
<EMBED>">="     { return(GREATEREQUAL); }
<EMBED>"~"      { return(NOT); }
<EMBED>"!"      { return(CUT); }


<EMBED>"%".*$       ;    /* comment; ignore */

<EMBED>{whitespace}+ ;    /* space/tab/formfeed; ignore */

<EMBED>\n|\r   {
                 yylineno++;
                 return(EOLN);
               }

<EMBED>.       {      /* any other single char; error */
                 fprintf(error_file,
                           "Error: Line %d: Illegal Character",
                           yylineno);
                 if (yytext[0]>=' ' && yytext[0] <='~')
                   fprintf(error_file,": `%c'\n", yytext[0]);
                 else
                   fprintf(error_file,
                           ", ASCII: %03o (octal)\n", yytext[0]);
               }


<EX_PROTO>.    {
                 fprintf(stderr,"In EX_PROTO, no match for '%c'\n",
                   *yytext);
                 PUT;
               }


<<EOF>>        {
                 return(EOF_TOK);
               }

%%

/************************************************************************
//
// ttwrap() - This function is called by the lexer [ttlex()] when
//            the end-of-file is reached.
//
/***********************************************************************/

int ttwrap()
{
    if (in_comment) {
        fprintf(error_file,
            "ERROR: End of file reached inside comment.\n" );
        in_comment = 0;
    }

    return(1);
}

/************************************************************************/
/* This eases the burden of debugging a file that is translated */

void line_directive(FILE *out)
{
    fprintf(out,"\n#line %d \"%s\"\n", yylineno, input_file_name);
}

/************************************************************************/
