// *************************************************************************
//    MakeTuple.C
// *************************************************************************

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "MakeTuple.h"

#include "HandleExport.h"
#include "MyString.h"
#include "HashDefines.h"
#include "Translate.h"
#include "OutputExport.h"
#include "Lexer.h"
#include "Token.h"
#include "Main.h"

/*****************************************************************************/

extern int coral_var_number;

/*****************************************************************************/
char* gen_tuple_stmt(struct vars_list *rel) 
{
    char *tuple_name;

    if (rel) {
        tuple_name = new_name();
        fprintf(stderr, "corTuple *%s =  new corTuple(", tuple_name);
        if (rel->arg_count < 5) 
            fprintf(stderr, "make_arglist%d(", rel->arg_count);
        else
            fprintf(stderr, "make_arglist(%d,", rel->arg_count);
        coral_var_number = 0;
        struct arg_list *args;
        args = rel->args;
        for (int i = 0; i < rel->arg_count - 1; i++) {
            gen_arg(args);
            fputs(",",stderr);
            args = args->next;
        }
        gen_arg(args);
        fputs("));\n",stderr);
    }
    else {
        fprintf(stderr, "rel is NULL \n");
    }
    
    return tuple_name;
}

/*****************************************************************************/
void gen_arg(struct arg_list *arg)
{
    char *s;

    if (arg == NULL)
        return;

    switch ((arg->arg_type)->kind) {

    case STRING:
    case IDENT:
    case EXPR:
        fprintf(stderr, "make_arg(%s)", (arg->arg_type)->val.str);
        break;
    case VAR:
        s = (arg->arg_type)->val.str;
        fprintf(stderr, "make_var(%s, %d)", s, coral_var_number++);
        break;
    case INUM: 
        fprintf(stderr, "make_arg(%d)", (arg->arg_type)->val.inum);
        break;
    case RNUM:
        fprintf(stderr, "make_arg(%f)", (arg->arg_type)->val.rnum);
        break;
    default:
        fprintf(stderr, "unknown argument in relation \n");
        break;
    }
}

/*****************************************************************************/
/*
    Procedure to handle decls like make_tuple(X,Y) etc. 
    This function gets each argument in the make tuple call by calling
    get_arg() function.  Then code to convert each argument to Arg * type
    is generated, through the means of make_arg and make_var defined in
    interface.C.
*/

void handle_make_tuple(int current_position, int /*end_position*/, char *str)
{
    int current;
    Token tokn, *t;
    current = get_token(&tokn);
    
    if (current != LPAREN) {
        fprintf(error_file, "syntax error \"(\" expected \n");
    }
    
    t = get_arg();
    
    struct arg_list *begin;
    struct arg_list *temp = new (struct arg_list);
    struct arg_list *next_temp;
    int arg_count = 0;
    
    begin = temp;
    
    while (current != RPAREN) {
        arg_count++;
        next_temp = new (struct arg_list);    
        next_temp->arg_type = t;
        temp->next = next_temp;
        temp = next_temp;
        t = get_arg();
        if (t->kind != RPAREN) {
            if (t->kind != COMMA) {
                int line_number = 0;
                int col_number = 0;
                fprintf(error_file,
                    "line %d col %d syntax error \",\" expected \n",
                    line_number, col_number);
            }
            t = get_arg();
        }
    }
    
    temp = begin;
    begin = begin->next;
    delete(temp);
    
    // begin points to beginning of arg_list
        
    char *name_of_tuple = output_make_tuple(arg_count, begin,
        str, current_position, 0);
}

/*****************************************************************************/
char* output_make_tuple(int arg_count, struct arg_list *begin,
            char *str, int pos1, int pos2)
{
    char buf[256];
    int pos = 0;
    int i;
    int len;
    struct arg_list *temp;
    char *name;
    
    char * name_of_tuple = new_name();
    fprintf(output_file, "corTuple * %s; \n{\n", name_of_tuple);
    for (i = 0, temp = begin; i < arg_count; i++, temp = temp->next) {
        gen_tuple_arg_stmt(output_file, i, temp);
    }
    
    len = strlen(name_of_tuple);
    strcpy(buf+pos, name_of_tuple);
    pos += len;
    buf[pos++] = '='; 
    name = "new corTuple(make_arglist(";
    len = strlen(name);
    strcpy(buf+pos, name);
    pos += len;
    
    if (arg_count != 0) {
        char tmp[10];
        sprintf(tmp, "%d", arg_count);
        len = strlen(tmp);
        strncpy(buf + pos, tmp, len);
        pos += len;
        buf[pos++] = ',';
    }
    
    for (i = 0; i < arg_count - 1; i++) {
        name = concat1("arg", i);
        len = strlen(name);
        strcpy(buf+pos, name);
        pos += len;
        buf[pos++] = ',';
    }
    if (arg_count != 0) {
        name = concat1("arg", arg_count-1);
        len = strlen(name);
        strcpy(buf+pos, name);
        pos += len;
    }

    strcpy(&(buf[pos]),"));");
   
    fputs(buf,output_file); 
    fputs("\n}\n", output_file);
    char *new_str = subs_str(str, name_of_tuple, pos1, pos2);
    fprintf(output_file, "%s\n", new_str);
    return(name_of_tuple);
}

/*****************************************************************************/
void gen_tuple_arg_stmt(FILE *output_file, int i, struct arg_list *arg)
{
    char *name; 
    char *s;

    if (arg == NULL)
        return;

    name = concat1("arg", i);
    fprintf(output_file, "corArg *%s = ", name);

    switch ((arg->arg_type)->kind) {
    case IDENT:
        s = (arg->arg_type)->val.str;
        fprintf(output_file, "make_arg(%s)", s);
        break;
    case VAR:
        s = (arg->arg_type)->val.str;
        fprintf(output_file, "make_var(\"%s\", %d)", s+1, coral_var_number++);
        // s + 1 to remove the dollar
        break;
    case EXPR:
        s = (arg->arg_type)->val.str;
        fprintf(output_file, "make_arg(%s)", s);
        break;
    default:
        fprintf(output_file, "make_arg(1000)"); // just for compiling
        fprintf(error_file, "unknown argument in relation.\n");
        break;
    }

    fputs(";\n",output_file);
}

/*****************************************************************************/
char *subs_str(char *str, char *name_of_tuple, int start_pos, int end_pos)
{
    // fprintf(stderr, "name is %s, str is %s positions are %d %d \n", name_of_tuple, str, start_pos, end_pos);*/
    
    int len = strlen(str);
    int len_of_name = strlen(name_of_tuple);
    int i,j;
    
    if (str == NULL)
        return str;
    
    if (name_of_tuple == NULL)
        return str;
    
    if (start_pos >= len)
        return str;
    
    if (end_pos >= len)
        return str;
    
    if ((end_pos - start_pos) >= len_of_name) {
        for (i = start_pos, j = 0; j < len_of_name; i++, j++) 
            str[i] = name_of_tuple[j];
        for (i = start_pos + len_of_name; i < end_pos; i++) 
            str[i] = ' ';
        return str;
    }
    else {
        char *new_str = new (char [len + len_of_name]);
        for (i = 0; i < start_pos; i++) 
            new_str[i] = str[i];
        for (i = start_pos, j = 0; j < len_of_name; i++, j++) 
            new_str[i] = name_of_tuple[j];
        for (i = end_pos; i < len; i++) 
            new_str[i] = str[i];
        return new_str;
    }
}

/*****************************************************************************/
